# sphinx configuration

project = "plone.app.event"
copyright = "2012, Plone Foundation"

extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
]
master_doc = "index"

language = "en"

from pkg_resources import get_distribution  # noqa: E402


version = release = get_distribution(project).version



from Acquisition import aq_inner
from ComputedAttribute import ComputedAttribute
from plone.app.event import _
from plone.app.event.base import _prepare_range
from plone.app.event.base import expand_events
from plone.app.event.base import get_events
from plone.app.event.base import localized_now
from plone.app.event.base import RET_MODE_ACCESSORS
from plone.app.event.base import start_end_query
from plone.app.event.portlets import get_calendar_url
from plone.app.event.portlets.portlet_calendar import ISyndicatableCollection
from plone.app.event.portlets.portlet_calendar import search_base_uid_source
from plone.app.portlets.portlets import base
from plone.app.querystring import queryparser
from plone.app.uuid.utils import uuidToObject
from plone.base.interfaces.controlpanel import ISiteSchema
from plone.event.interfaces import IEvent
from plone.memoize import ram
from plone.memoize.compress import xhtml_compress
from plone.portlets.interfaces import IPortletDataProvider
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zExceptions import NotFound
from zope import schema
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.contentprovider.interfaces import IContentProvider
from zope.interface import implementer


def _render_events_cachekey(method, self):
    # Since the events portlet shows the upcoming events, we simply get the
    # most recently modified event and return its uuid and last modified date
    # as key. If there is a change, then the most expensive computation can be
    # performed.

    site = getSite()
    cat = getToolByName(site, "portal_catalog")

    query = {}
    data = self.data
    if data.state:
        query["review_state"] = data.state

    start = localized_now(site)

    query.update(self.request.get("contentFilter", {}))
    if ISyndicatableCollection and ISyndicatableCollection.providedBy(self.search_base):
        # Whatever sorting is defined, we're overriding it.
        query = queryparser.parseFormquery(
            self.search_base, self.search_base.query, sort_on="start", sort_order=None
        )
        if "start" in query:
            start = query["start"]
    elif self.search_base_path:
        query["path"] = {"query": self.search_base_path}

    # Any object with an IEvent interface.
    query["object_provides"] = IEvent.__identifier__
    # Only upcoming events.
    query.update(start_end_query(start=start, end=None))
    # We want to get the next data.count events, just as the portlet would.
    query["sort_on"] = "start"
    query["sort_order"] = "asc"
    query["sort_limit"] = data.count

    events = cat(**query)
    uuid = ""
    modified = 0
    for event in events:
        uuid += event.UID
        modified += event.modified.asdatetime().timestamp()

    return (uuid, modified)


class IEventsPortlet(IPortletDataProvider):
    count = schema.Int(
        title=_("Number of items to display"),
        description=_("How many items to list."),
        required=True,
        default=5,
        min=1,
    )

    state = schema.Tuple(
        title=_("Workflow state"),
        description=_("Items in which workflow state to show."),
        default=None,
        required=False,
        value_type=schema.Choice(vocabulary="plone.app.vocabularies.WorkflowStates"),
    )

    search_base_uid = schema.Choice(
        title=_("portlet_label_search_base", default="Search base"),
        description=_(
            "portlet_help_search_base",
            default="Select search base Folder or Collection to search for "
            "events. The URL to to this item will also be used to "
            "link to in calendar searches. If empty, the whole site "
            "will be searched and the event listing view will be "
            "called on the site root.",
        ),
        required=False,
        source=search_base_uid_source,
    )
    thumb_scale = schema.TextLine(
        title=_("Override thumb scale"),
        description=_(
            "Enter a valid scale name"
            " (see 'Image Handling' control panel) to override"
            " (e.g. icon, tile, thumb, mini, preview, ... )."
            " Leave empty to use default (see 'Site' control panel)."
        ),
        required=False,
        default="",
    )

    no_thumbs = schema.Bool(
        title=_("Suppress thumbs"),
        description=_("If enabled, the portlet will not show thumbs."),
        required=False,
        default=False,
    )


@implementer(IEventsPortlet)
class Assignment(base.Assignment):
    # reduce upgrade pain
    thumb_scale = None
    no_thumbs = False

    def __init__(
        self,
        count=5,
        state=None,
        search_base_uid=None,
        thumb_scale=None,
        no_thumbs=False,
    ):
        self.count = count
        self.state = state
        self.search_base_uid = search_base_uid
        self.thumb_scale = thumb_scale
        self.no_thumbs = no_thumbs

    @property
    def title(self):
        return _("Events")

    def _uid(self):
        # This is only called if the instance doesn't have a search_base_uid
        # attribute, which is probably because it has an old
        # 'search_base' attribute that needs to be converted.
        path = self.search_base
        try:
            search_base = getSite().unrestrictedTraverse(path.lstrip("/"))
        except (AttributeError, KeyError, TypeError, NotFound):
            return
        return search_base.UID()

    search_base_uid = ComputedAttribute(_uid, 1)


class Renderer(base.Renderer):
    _template = ViewPageTemplateFile("portlet_events.pt")
    _search_base = None

    @property
    def search_base(self):
        if not self._search_base and self.data.search_base_uid:
            self._search_base = uuidToObject(self.data.search_base_uid)
        return aq_inner(self._search_base) if self._search_base else None

    @property
    def search_base_path(self):
        return (
            "/".join(self.search_base.getPhysicalPath()) if self.search_base else None
        )  # noqa

    def update(self):
        context = aq_inner(self.context)
        calendar_url = get_calendar_url(context, self.search_base_path)
        self.next_url = "%s?mode=future" % calendar_url
        self.prev_url = "%s?mode=past" % calendar_url

    def render(self):
        return xhtml_compress(self._template())

    @property
    def available(self):
        return self.data.count > 0 and len(self.events)

    @property
    @ram.cache(_render_events_cachekey)
    def events(self):
        context = aq_inner(self.context)
        data = self.data

        query = {}
        if data.state:
            query["review_state"] = data.state

        events = []
        query.update(self.request.get("contentFilter", {}))
        if ISyndicatableCollection.providedBy(self.search_base):
            # Whatever sorting is defined, we're overriding it.
            query = queryparser.parseFormquery(
                self.search_base,
                self.search_base.query,
                sort_on="start",
                sort_order=None,
            )

            start = None
            if "start" in query:
                start = query["start"]
            else:
                start = localized_now(context)

            end = None
            if "end" in query:
                end = query["end"]

            start, end = _prepare_range(self.search_base, start, end)
            query.update(start_end_query(start, end))
            events = self.search_base.results(
                batch=False, brains=True, custom_query=query, limit=data.count
            )
            events = expand_events(
                events,
                ret_mode=RET_MODE_ACCESSORS,
                start=start,
                end=end,
                sort="start",
                sort_reverse=False,
            )
            events = events[: data.count]  # limit expanded
        else:
            if self.search_base_path:
                query["path"] = {"query": self.search_base_path}
            events = get_events(
                context,
                start=localized_now(context),
                ret_mode=RET_MODE_ACCESSORS,
                expand=True,
                limit=data.count,
                **query,
            )

        return events

    def formatted_date(self, event):
        provider = getMultiAdapter(
            (self.context, self.request, self), IContentProvider, name="formatted_date"
        )
        return provider(event)

    def thumb_scale(self):
        """Use override value or read thumb_scale from registry.
        Image sizes must fit to value in allowed image sizes.
        None will suppress thumb.
        """
        if getattr(self.data, "no_thumbs", False):
            # Individual setting overrides ...
            return None
        thsize = getattr(self.data, "thumb_scale", None)
        if thsize:
            return thsize
        registry = getUtility(IRegistry)
        settings = registry.forInterface(ISiteSchema, prefix="plone", check=False)
        if settings.no_thumbs_portlet:
            return None
        thumb_scale_portlet = settings.thumb_scale_portlet
        return thumb_scale_portlet


class AddForm(base.AddForm):
    schema = IEventsPortlet
    label = _("Add Events Portlet")
    description = _("This portlet lists upcoming Events.")

    def create(self, data):
        return Assignment(
            count=data.get("count", 5),
            state=data.get("state", None),
            search_base_uid=data.get("search_base_uid", None),
        )


class EditForm(base.EditForm):
    schema = IEventsPortlet
    label = _("Edit Events Portlet")
    description = _("This portlet lists upcoming Events.")



from Acquisition import aq_parent
from OFS.SimpleItem import SimpleItem
from plone.app.event.base import dt_end_of_day
from plone.app.event.base import dt_start_of_day
from plone.app.event.base import guess_date_from
from plone.base.utils import safe_text
from plone.event.interfaces import IEventAccessor
from plone.event.interfaces import IEventRecurrence
from plone.event.interfaces import IOccurrence
from plone.event.interfaces import IRecurrenceSupport
from plone.event.recurrence import recurrence_sequence_ical
from plone.event.utils import is_same_day
from plone.event.utils import pydt
from plone.namedfile.interfaces import IImageScaleTraversable
from plone.namedfile.scaling import ImageScaling
from Products.Five.browser import BrowserView
from zope.component import adapter
from zope.interface import implementer
from zope.publisher.interfaces.browser import IBrowserRequest
from ZPublisher.BaseRequest import DefaultPublishTraverse


@adapter(IEventRecurrence)
@implementer(IRecurrenceSupport)
class RecurrenceSupport:
    """IRecurrenceSupport Adapter."""

    def __init__(self, context):
        self.context = context

    def occurrences(self, range_start=None, range_end=None):
        """Return all occurrences of an event, possibly within a start and end
        limit.

        :param range_start: Optional start datetime, from which you want
                            occurrences be returned.
        :type range_start: Python datetime
        :param range_end: Optional start datetime, from which you want
                          occurrences be returned.
        :type range_end: Python datetime
        :returns: List of occurrences, including the start event.
        :rtype: IEvent or IOccurrence based objects

        Please note: Events beginning before range_start but ending afterwards
                     won't be found.

        TODO: really?

        TODO: test with event start = 21st feb, event end = start+36h,
        recurring for 10 days, range_start = 1st mar, range_end = last Mark
        """
        event = IEventAccessor(self.context)

        # We try to get IEventBasic start without including recurrence
        event_start = getattr(self.context, "start", None)
        if not event_start:
            event_start = event.start
        elif getattr(event, "whole_day", None):
            event_start = dt_start_of_day(event_start)

        # We get event ends by adding a duration to the start. This way, we
        # prevent that the start and end lists are of different size if an
        # event starts before range_start but ends afterwards.
        if getattr(event, "open_end", None):
            event_end = dt_end_of_day(event_start)
        elif getattr(event, "whole_day", None):
            event_end = dt_end_of_day(getattr(self.context, "end", None))
        else:
            event_end = getattr(self.context, "end", None)

        duration = event_end - event_start

        starts = recurrence_sequence_ical(
            event_start,
            recrule=event.recurrence,
            from_=range_start,
            until=range_end,
            duration=duration,
        )

        # XXX potentially occurrence won't need to be wrapped anymore
        # but doing it for backwards compatibility as views/templates
        # still rely on acquisition-wrapped objects.
        def get_obj(start):
            if pydt(event_start.replace(microsecond=0)) == start:
                # If the occurrence date is the same as the event object, the
                # occurrence is the event itself. return it as such.
                # Dates from recurrence_sequence_ical are explicitly without
                # microseconds, while event.start may contain it. So we have to
                # remove it for a valid comparison.
                return self.context
            return Occurrence(
                id=str(start.date()), start=start, end=start + duration
            ).__of__(self.context)

        for start in starts:
            yield get_obj(start)


@adapter(IEventRecurrence, IBrowserRequest)
class OccurrenceTraverser(DefaultPublishTraverse):
    """Generic Occurrence traverser.

    Please note: the .at and .dx subpackages implement their own Occurrence
    traversers.
    """

    def publishTraverse(self, request, name):
        context = self.context
        dateobj = guess_date_from(name, context)
        if dateobj:
            occs = IRecurrenceSupport(context).occurrences(range_start=dateobj)
            try:
                occurrence = next(occs)
                occ_acc = IEventAccessor(occurrence)
                if is_same_day(dateobj, occ_acc.start):
                    return occurrence
            except StopIteration:
                pass
        return self.fallbackTraverse(request, name)

    def fallbackTraverse(self, request, name):
        return super().publishTraverse(request, name)


@implementer(IOccurrence)
class Occurrence(SimpleItem):
    """Transient Occurrence object, representing an individual event in a
    recurrecne set.
    """

    def __init__(self, id, start, end):
        self.id = id
        self.start = start
        self.end = end
        self.portal_type = "Occurrence"


@adapter(IOccurrence)
@implementer(IEventAccessor)
class EventOccurrenceAccessor:
    """Generic event accessor adapter implementation for Occurrence objects."""

    def __init__(self, context):
        object.__setattr__(self, "context", context)

        own_attr = ["start", "end", "url"]
        object.__setattr__(self, "_own_attr", own_attr)

    def _get_context(self, name):
        # TODO: save parent context on self, so it must not be called every
        #       time
        oa = self._own_attr
        if name in oa:
            return self.context
        else:
            return IEventAccessor(aq_parent(self.context))

    def __getattr__(self, name):
        return getattr(self._get_context(name), name, None)

    def __setattr__(self, name, value):
        return setattr(self._get_context(name), name, value)

    def __delattr__(self, name):
        delattr(self._get_context(name), name)

    # R/O properties
    # TODO: Having uid here makes probably no sense, since Occurrences are
    #       created on the fly and not persistent.

    @property
    def url(self):
        return safe_text(self.context.absolute_url())


class ImageScalingViewFactory(BrowserView):
    """Factory for ImageScaling view for Occurrences.
    Delegates to parent @@images view.
    """

    def __new__(cls, context, request):
        parent = aq_parent(context)
        if IImageScaleTraversable.providedBy(parent):
            return ImageScaling(parent, request)
        return None



from Acquisition import aq_inner
from Acquisition import aq_parent
from calendar import monthrange
from datetime import date
from datetime import datetime
from datetime import timedelta
from DateTime import DateTime
from persistent.dict import PersistentDict
from plone.app.event.interfaces import ISO_DATE_FORMAT
from plone.base.i18nl10n import ulocalized_time as orig_ulocalized_time
from plone.base.interfaces import INavigationRoot
from plone.base.interfaces.siteroot import IPloneSiteRoot
from plone.base.navigationroot import get_navigation_root_object
from plone.base.utils import safe_callable
from plone.event.interfaces import IEvent
from plone.event.interfaces import IEventAccessor
from plone.event.interfaces import IEventRecurrence
from plone.event.interfaces import IRecurrenceSupport
from plone.event.utils import default_timezone as fallback_default_timezone
from plone.event.utils import dt2int
from plone.event.utils import is_date
from plone.event.utils import is_datetime
from plone.event.utils import is_same_day
from plone.event.utils import is_same_time
from plone.event.utils import pydt
from plone.event.utils import validated_timezone
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from zope.annotation.interfaces import IAnnotations
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.component.interfaces import ISite
from zope.deprecation import deprecate

import pytz


DEFAULT_END_DELTA = 1  # hours
FALLBACK_TIMEZONE = "UTC"

# Sync strategies
SYNC_NONE = 0
SYNC_KEEP_NEWER = 1
SYNC_KEEP_MINE = 2
SYNC_KEEP_THEIRS = 3

# Return modes for get_events
RET_MODE_BRAINS = 1
RET_MODE_OBJECTS = 2
RET_MODE_ACCESSORS = 3

# Map for ambiguous timezone abbreviations to their most common non-ambigious
# timezone name. E.g CST is ambiguous and is used for U.S./Canada Central
# Standard Time, Australian Central Standard Time, China Standard Time.
# TODO: incomplete map.
# TODO: do we need this at all or shouldn't we just fail with ambiguous
#       timezones?
replacement_zones = {
    "CET": "Europe/Vienna",  # Central European Time
    "MET": "Europe/Vienna",  # Middle European Time
    "EET": "Europe/Helsinki",  # East European Time
    "WET": "Europe/Lisbon",  # West European Time
}


# RETRIEVE EVENTS


def get_events(
    context,
    start=None,
    end=None,
    limit=None,
    ret_mode=RET_MODE_BRAINS,
    expand=False,
    sort="start",
    sort_reverse=False,
    **kw,
):
    """Return all events as catalog brains, possibly within a given
    timeframe.

    :param context: [required] A context object.
    :type context: Content object

    :param start: Date, from which on events should be searched.
    :type start: Python datetime.

    :param end: Date, until which events should be searched.
    :type end: Python datetime

    :param limit: Number of items to be returned.
    :type limit: integer

    :param ret_mode: Return type of search results. These options are
                     available:

                         * 1 (brains): Return results as catalog brains.
                         * 2 (objects): Return results as IEvent and/or
                                        IOccurrence objects.
                         * 3 (accessors): Return results as IEventAccessor
                                          wrapper objects.
    :type ret_mode: integer [1|2|3]

    :param expand: Expand the results to all occurrences (within a timeframe,
                   if given). With this option set to True, the resultset also
                   includes the event's recurrence occurrences and is sorted by
                   the start date.
                   Only available in ret_mode 2 (objects) and 3 (accessors).
    :type expand: boolean

    :param sort: Catalog index id to sort on.
    :type sort: string

    :param sort_reverse: Change the order of the sorting.
    :type sort_reverse: boolean

    :returns: Portal events, matching the search criteria.
    :rtype: catalog brains, event objects or IEventAccessor object wrapper,
            depending on ret_mode.
    """
    start, end = _prepare_range(context, start, end)

    query = {}
    query["object_provides"] = IEvent.__identifier__

    query.update(start_end_query(start, end))

    if "path" not in kw:
        # limit to the current navigation root, usually (not always) site
        portal = getSite()
        navroot = get_navigation_root_object(context, portal)
        query["path"] = "/".join(navroot.getPhysicalPath())
    else:
        query["path"] = kw["path"]

    # Sorting
    # In expand mode we sort after calculation of recurrences again. But we
    # need to leave this sorting here in place, since no sort definition could
    # lead to arbitrary results when limiting with sort_limit.
    query["sort_on"] = sort
    if sort_reverse:
        query["sort_order"] = "reverse"

    # cannot limit before resorting or expansion, see below

    query.update(kw)

    cat = getToolByName(context, "portal_catalog")
    result = cat(**query)

    # unfiltered catalog results are already sorted correctly on brain.start
    # filtering on start/end requires a resort, see docstring below and
    # p.a.event.tests.test_base_module.TestGetEventsDX.test_get_event_sort
    if sort in ("start", "end"):
        result = filter_and_resort(context, result, start, end, sort, sort_reverse)

        # Limiting a start/end-sorted result set is possible here
        # and provides an important optimization BEFORE costly expansion
        if limit:
            result = result[:limit]

    if ret_mode in (RET_MODE_OBJECTS, RET_MODE_ACCESSORS):
        if expand is False:
            result = [_obj_or_acc(it.getObject(), ret_mode) for it in result]
        else:
            result = expand_events(result, ret_mode, start, end, sort, sort_reverse)

    # Limiting a non-start-sorted result set can only happen here
    if limit:
        result = result[:limit]

    return result


def filter_and_resort(context, brains, start, end, sort, sort_reverse):
    """#114 sorting bug is fallout from a Products.DateRecurringIndex
    limitation. The index contains a set of start and end dates
    represented as integer: that allows valid slicing of searches.
    However the returned brains have a .start attribute which is
    the start DateTime of the *first* occurrence of an event.

    This results in mis-sorting of search results if the next occurrence
    of event B is after the next occurrence of event A, but the first
    occurrence of event B is *before* the first occurrence of event A.
    The catalog results sort that as B<A instead of A<B.

    This method works around that issue by extracting all occurrence
    start/end from the index, and then sorting on the actual next start/end.

    For ongoing events which have an occurrence starting in the past
    but ending in the future, the past start of that ongoing occurrence
    is selected, so this will show up right at the start of the result.

    :param context: [required] A context object.
    :type context: Content object

    :param brains: [required] catalog brains
    :type brains: catalog brains

    :param start: [required] min end datetime (sic!)
    :type start: Python datetime.

    :param end: [required] max start datetime (sic!)
    :type start: Python datetime.

    :param sort_reverse: Change the order of the sorting.
    :type sort_reverse: boolean

    :param sort: Which field to sort on
    :type sort: 'start' or 'end'

    :returns: catalog brains
    :rtype: catalog brains

    """
    _start = dt2int(start)  # index contains longint sets
    _end = dt2int(end)
    catalog = getToolByName(context, "portal_catalog")
    items = []  # (start:int, occurrence:brain) pairs
    for brain in brains:
        # brain.start metadata reflects first occurrence.
        # instead, get all occurrence start/end from raw index
        idx = catalog.getIndexDataForRID(brain.getRID())
        _allstarts = sorted(idx["start"])
        _allends = sorted(idx["end"])
        # assuming (start, end) pairs belong together
        # assert(len(_allstarts) == len(_allends))
        _occ = zip(_allstarts, _allends)
        if start:
            _occ = [(s, e) for (s, e) in _occ if e >= _start]
        if end:
            _occ = [(s, e) for (s, e) in _occ if s <= _end]
        if not _occ:
            continue
        if sort == "start":
            # first start can be before filter window if end is in window
            _first = min(s for (s, e) in _occ)
        elif sort == "end":
            _first = min(e for (s, e) in _occ)
        items.append((_first, brain))  # key on next start/end

    # sort brains by next start, discard sort key
    data = [x[1] for x in sorted(items, key=lambda x: x[0])]
    if sort_reverse:
        data.reverse()
    return data


def expand_events(events, ret_mode, start=None, end=None, sort=None, sort_reverse=None):
    """Expand to the recurrence occurrences of a given set of events.

    :param events: IEvent based objects or IEventAccessor object wrapper.

    :param ret_mode: Return type of search results. These options are
                     available:

                         * 2 (objects): Return results as IEvent and/or
                                        IOccurrence objects.
                         * 3 (accessors): Return results as IEventAccessor
                                          wrapper objects.
                     Option "1" (brains) is not supported.

    :type ret_mode: integer [2|3]

    :param start: Date, from which on events should be expanded.
    :type start: Python datetime.

    :param end: Date, until which events should be expanded.
    :type end: Python datetime

    :param sort: Object or IEventAccessor Attribute to sort on.
    :type sort: string

    :param sort_reverse: Change the order of the sorting.
    :type sort_reverse: boolean
    """
    assert ret_mode is not RET_MODE_BRAINS

    exp_result = []
    for it in events:
        obj = it.getObject() if getattr(it, "getObject", False) else it
        if IEventRecurrence.providedBy(obj):
            occ_list = list(IRecurrenceSupport(obj).occurrences(start, end))
            # add original event (to make expand_events not remove one item of the original list)
            if (start or end) and (obj.start >= (end or start)):
                if obj not in occ_list:
                    occ_list.append(obj)
            occurrences = [_obj_or_acc(occ, ret_mode) for occ in occ_list]
        elif IEvent.providedBy(obj):
            occurrences = [_obj_or_acc(obj, ret_mode)]
        else:
            # No IEvent based object. Could come from a collection.
            continue
        exp_result += occurrences
    if sort:
        exp_result.sort(key=lambda x: _get_compare_attr(x, sort))
    if sort_reverse:
        exp_result.reverse()
    return exp_result


def _obj_or_acc(obj, ret_mode):
    """Return the content object or an IEventAccessor wrapper, depending on
    ret_mode. ret_mode 2 returns objects, ret_mode 3 returns IEventAccessor
    object wrapper. ret_mode 1 is not supported.
    """
    assert ret_mode is not RET_MODE_BRAINS
    if ret_mode == RET_MODE_OBJECTS:
        return obj
    elif ret_mode == RET_MODE_ACCESSORS:
        return IEventAccessor(obj)


def _get_compare_attr(obj, attr):
    """Return an compare attribute, supporting AT, DX and IEventAccessor
    objects.
    """
    val = getattr(obj, attr, None)
    if safe_callable(val):
        val = val()
    if isinstance(val, DateTime):
        val = pydt(val)
    return val


def construct_calendar(events, start=None, end=None):
    """Return a dictionary with dates in a given timeframe as keys and the
    actual occurrences for that date for building calendars.
    Long lasting events will occur on every day until their end.

    :param events: List of IEvent and/or IOccurrence objects, to construct a
                   calendar data structure from.
    :type events: list

    :param start: An optional start range date.
    :type start: Python datetime or date

    :param end: An optional start range date.
    :type end: Python datetime or date

    :returns: Dictionary with isoformat date strings as keys and event
              occurrences as values.
    :rtype: dict

    """
    if start:
        if is_datetime(start):
            start = start.date()
        assert is_date(start)
    if end:
        if is_datetime(end):
            end = end.date()
        assert is_date(end)

    cal = {}

    def _add_to_cal(cal_data, event, date):
        date_str = date.isoformat()
        if date_str not in cal_data:
            cal_data[date_str] = [event]
        else:
            cal_data[date_str].append(event)
        return cal_data

    for event in events:
        acc = IEventAccessor(event)
        start_date = acc.start.date()
        end_date = acc.end.date()

        # day span between start and end + 1 for the initial date
        range_days = (end_date - start_date).days + 1
        for add_day in range(range_days):
            next_start_date = start_date + timedelta(add_day)  # initial = 0

            # avoid long loops
            if start and end_date < start:
                break  # if the date is completely outside the range
            if start and next_start_date < start:
                continue  # if start_date is outside but end reaches into range
            if end and next_start_date > end:
                break  # if date is outside range

            _add_to_cal(cal, event, next_start_date)
    return cal


def _prepare_range(context, start, end):
    """Prepare a date-range to contain timezone info and set end to next day,
    if end is a date.

    :param context: [required] Context object.
    :type context: Content object

    :param start: [required] Range start.
    :type start: Python date or datetime

    :param end: [required] Range end.
    :type end: Python date or datetime

    :returns: Localized start and end datetime.
    :rtype: tuple

    """
    tz = default_timezone(context, as_tzinfo=True)
    start = pydt(start, missing_zone=tz)
    if is_date(end):
        # set range_end to the next day, time will be 0:00
        # so the whole previous day is also used for search
        end = end + timedelta(days=1)
    end = pydt(end, missing_zone=tz)
    return start, end


def start_end_query(start, end):
    """Make a catalog query out of start and end dates."""
    query = {}
    if start:
        # All events from start date ongoing:
        # The minimum end date of events is the date from which we search.
        query["end"] = {"query": start, "range": "min"}
    if end:
        # All events until end date:
        # The maximum start date must be the date until we search.
        query["start"] = {"query": end, "range": "max"}
    return query


# TIMEZONE HANDLING


def default_timezone(context=None, as_tzinfo=False):
    """Return the timezone from the portal or user.

    :param context: Optional context. If not given, the current Site is used.
    :type context: Content object

    :param as_tzinfo: Return the default timezone as tzinfo object.
    :type as_tzinfo: boolean

    :returns: Timezone identifier or tzinfo object.
    :rtype: string or tzinfo object

    """
    # TODO: test member timezone
    if not context:
        context = getSite()

    membership = getToolByName(context, "portal_membership", None)
    if membership and not membership.isAnonymousUser():  # user not logged in
        member = membership.getAuthenticatedMember()
        member_timezone = member.getProperty("timezone", None)
        if member_timezone:
            info = pytz.timezone(member_timezone)
            return info if as_tzinfo else info.zone

    reg_key = "plone.portal_timezone"
    registry = getUtility(IRegistry)
    portal_timezone = registry.get(reg_key, None)

    # fallback to what plone.event is doing
    if not portal_timezone:
        portal_timezone = fallback_default_timezone()

    # Change any ambiguous timezone abbreviations to their most common
    # non-ambigious timezone name.
    if portal_timezone in replacement_zones.keys():
        portal_timezone = replacement_zones[portal_timezone]
    portal_timezone = validated_timezone(portal_timezone, FALLBACK_TIMEZONE)

    if as_tzinfo:
        return pytz.timezone(portal_timezone)

    return portal_timezone


def localized_now(context=None):
    """Return the current datetime localized to the default timezone.

    :param context: Context object.
    :type context: Content object
    :returns: Localized current datetime.
    :rtype: Python datetime

    """
    if not context:
        context = getSite()
    tzinfo = default_timezone(context=context, as_tzinfo=True)
    return datetime.now(tzinfo).replace(microsecond=0)


def localized_today(context=None):
    """Return the current date localized to the default timezone.

    :param context: Context object.
    :type context: Content object
    :returns: Localized current date.
    :rtype: Python date

    """
    now = localized_now(context)
    return date(now.year, now.month, now.day)


# DATETIME HELPERS


def first_weekday():
    """Returns the number of the first Weekday in a Week, as defined in
    the registry. 0 is Monday, 6 is Sunday, as expected by Python's datetime.

    PLEASE NOTE: strftime %w interprets 0 as Sunday unlike the calendar module!

    :returns: Index of first weekday [0(Monday)..6(Sunday)]
    :rtype: integer

    """
    reg_key = "plone.first_weekday"
    registry = getUtility(IRegistry)
    first_wd = registry.get(reg_key, None)

    if not first_wd:
        return 0
    else:
        return int(first_wd)


def wkday_to_mon0(day):
    """Converts an integer weekday number to a representation where Monday is 0
    and Sunday is 6 (the datetime default), from a representation where Sunday
    is 0, Monday is 1 and Saturday is 6 (the strftime behavior).

    :param day: The weekday number [0(Sunday)..6]
    :type day: integer

    :returns: The weekday number [0(Monday)..6]
    :rtype: integer

    """
    if day == 0:
        return 6
    else:
        return day - 1


def wkday_to_mon1(day):
    """Converts an integer weekday number to a representation where Monday is
    1, Saturday is 6 and Sunday is 0 (the strftime behavior), from a
    representation where Monday is 0 and Sunday is 6 (the datetime default).

    :param day: The weekday number [0(Monday)..6]
    :type day: integer

    :returns: The weekday number [0(Sunday)..6]
    :rtype: integer

    """
    if day == 6:
        return 0
    else:
        return day + 1


def DT(dt, exact=False):
    """Return a Zope DateTime instance from a Python datetime instance.

    :param dt: Python datetime, Python date, Zope DateTime instance or string.
    :param exact: If True, the resolution goes down to microseconds. If False,
                  the resolution are seconds. Default is False.
    :type exact: Boolean
    :returns: Zope DateTime
    :rtype: Zope DateTime

    """

    def _adjust_DT(DT, exact):
        if exact:
            ret = DT
        else:
            ret = DateTime(
                DT.year(),
                DT.month(),
                DT.day(),
                DT.hour(),
                DT.minute(),
                int(DT.second()),
                DT.timezone(),
            )
        return ret

    tz = default_timezone(getSite())
    ret = None
    if is_datetime(dt):
        zone_id = getattr(dt.tzinfo, "zone", tz)
        tz = validated_timezone(zone_id, tz)
        second = dt.second
        if exact:
            second += dt.microsecond / 1000000.0
        ret = DateTime(dt.year, dt.month, dt.day, dt.hour, dt.minute, second, tz)
    elif is_date(dt):
        ret = DateTime(dt.year, dt.month, dt.day, 0, 0, 0, tz)
    elif isinstance(dt, DateTime):
        # No timezone validation. DateTime knows how to handle it's zones.
        ret = _adjust_DT(dt, exact=exact)
    else:
        # Try to convert by DateTime itself
        ret = _adjust_DT(DateTime(dt), exact=exact)
    return ret


def guess_date_from(datestr, context=None):
    """Returns a timezone aware date object if an arbitrary ASCII string is
    formatted in an ISO date format, otherwise None is returned.

    Used for traversing and Occurrence ids.

    :param datestr: Date string in an ISO format.
    :type datestr: string
    :param context: Context object (for retrieving the timezone).
    :type context: Content object
    :returns: Localized date object.
    :rtype: Python date

    """
    try:
        dateobj = datetime.strptime(datestr, ISO_DATE_FORMAT)
    except ValueError:
        return

    ret = pytz.timezone(default_timezone(context)).localize(dateobj)
    return ret


def dt_start_of_day(dt):
    """Returns a Python datetime instance set to the start time of the given
    day (00:00:00).

    :param dt: datetime to set to the start time of the day.
    :type dt: Python datetime
    :returns: datetime set to the start time of the day (00:00:00).
    :rtype: Python datetime

    """
    if not isinstance(dt, datetime):
        # is a date
        dt = datetime.fromordinal(dt.toordinal())
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def dt_end_of_day(dt):
    """Returns a Python datetime instance set to the end time of the given day
    (23:59:59).

    :param dt: datetime to set to the end time of the day.
    :type dt: Python datetime
    :returns: datetime set to the end time of the day (23:59:59).
    :rtype: Python datetime

    """
    if not isinstance(dt, datetime):
        # is a date
        dt = datetime.fromordinal(dt.toordinal())
    return dt.replace(hour=23, minute=59, second=59, microsecond=0)


def start_end_from_mode(mode, dt=None, context=None):
    """Return a start and end date from a given mode string, like
    "today", "past" or "future". This can be used in event retrieval
    functions.

    :param mode: One of the following modes:

                    'all' Show all events.
                    'past': Show only past events with descending sorting.
                    'future': Show only future events (default).
                    'today': Show todays events.
                    'now': Show todays upcoming events.
                    '7days': Show events until 7 days in future.
                    'day': Return all events on the given day (dt parameter
                           required)
                    'week': Show a weeks events, optionally from a given date.

                 These settings override the start and end parameters.

                 Not implemented yet:

                    'month': Show this month's events.

    :type mode: string

    :param dt: Optional datetime for day mode.
    :type dt: Python datetime

    """
    if not context:
        context = getSite()
    now = localized_now(context)
    start = end = None

    if mode == "all":
        start = None
        end = None

    elif mode == "past":
        start = None
        end = now

    elif mode == "future":
        start = now
        end = None

    elif mode == "now":
        start = now
        end = dt_end_of_day(now)

    elif mode == "7days":
        start = now
        end = dt_end_of_day(now + timedelta(days=6))

    elif mode == "day" or mode == "today":
        if not dt:
            dt = now  # show today
        start = dt_start_of_day(dt)
        end = dt_end_of_day(dt)

    elif mode == "week":
        if not dt:
            dt = now  # show this week
        wkd = dt.weekday()
        first = first_weekday()

        if first <= wkd:
            delta = wkd - first  # >= 0
        if first > wkd:
            delta = wkd + 7 - first  # > 0

        start = dt_start_of_day(dt - timedelta(days=delta))
        end = dt_end_of_day(start + timedelta(days=6))

    elif mode == "month":
        if not dt:
            dt = now  # show this month
        year = dt.year
        month = dt.month
        last_day = monthrange(year, month)[1]  # (wkday, days)
        start = dt_start_of_day(datetime(year, month, 1))
        end = dt_end_of_day(datetime(year, month, last_day))

    return start, end


# DISPLAY HELPERS


def dates_for_display(occurrence):
    """Return a dictionary containing pre-calculated information for building
    <start>-<end> date strings.

    Keys are:
        'start_date' - date string of the start date
        'start_time' - time string of the start date
        'end_date'   - date string of the end date
        'end_time'   - time string of the end date
        'start_iso'  - start date in iso format
        'end_iso'    - end date in iso format
        'same_day'   - event ends on the same day
        'same_time'  - event ends at same time
        'whole_day'  - whole day events
        'open_end'   - events without end time

    :param occurrence: Event or occurrence object.
    :type occurrence: IEvent, IOccurrence or IEventAccessor based object.
    :returns: Dictionary with date strings.
    :rtype: dict


    The behavior os ulocalized_time() with time_only is odd.
    Setting time_only=False should return the date part only and *not*
    the time

    NOTE: these tests are not run, but serve as documentation.
    TODO: remove.
    >>> from DateTime import DateTime
    >>> start = DateTime(2010,3,16,14,40)
    >>> from zope.component.hooks import getSite
    >>> site = getSite()
    >>> ulocalized_time(start, False,  time_only=True, context=site)
    u'14:40'
    >>> ulocalized_time(start, False,  time_only=False, context=site)
    u'14:40'
    >>> ulocalized_time(start, False,  time_only=None, context=site)
    u'16.03.2010'

    """
    if IEventAccessor.providedBy(occurrence):
        acc = occurrence
        occurrence = occurrence.context
    else:
        acc = IEventAccessor(occurrence)

    if acc.start is None or acc.end is None:
        # Eventually optional start/end dates from a potentially Event.
        return None

    # this needs to separate date and time as ulocalized_time does
    DT_start = DT(acc.start)
    DT_end = DT(acc.end)
    start_date = ulocalized_time(
        DT_start, long_format=False, time_only=None, context=occurrence
    )
    start_time = ulocalized_time(
        DT_start, long_format=False, time_only=True, context=occurrence
    )
    end_date = ulocalized_time(
        DT_end, long_format=False, time_only=None, context=occurrence
    )
    end_time = ulocalized_time(
        DT_end, long_format=False, time_only=True, context=occurrence
    )

    same_day = is_same_day(acc.start, acc.end)
    same_time = is_same_time(acc.start, acc.end)

    # set time fields to None for whole day events
    if acc.whole_day:
        start_time = end_time = None
    if acc.open_end:
        end_time = None

    start_iso = acc.whole_day and acc.start.date().isoformat() or acc.start.isoformat()
    end_iso = acc.whole_day and acc.end.date().isoformat() or acc.end.isoformat()

    return dict(
        # Start
        start_date=start_date,
        start_time=start_time,
        start_iso=start_iso,
        # End
        end_date=end_date,
        end_time=end_time,
        end_iso=end_iso,
        # Meta
        same_day=same_day,
        same_time=same_time,
        whole_day=acc.whole_day,
        open_end=acc.open_end,
    )


@deprecate("date_speller is no longer supported, use spell_date instead.")
def date_speller(context, dt):
    return spell_date(dt, context)


def spell_date(dt, translation_context=None):
    """Return a dictionary with localized and readable formatted date parts."""
    if not translation_context:
        translation_context = getSite()

    dt = DT(dt)
    util = getToolByName(translation_context, "translation_service")
    dom = "plonelocales"

    def zero_pad(num):
        return "%02d" % num

    date_dict = dict(
        year=dt.year(),
        month=dt.month(),
        month2=zero_pad(dt.month()),
        month_name=util.translate(
            util.month_msgid(dt.month()), domain=dom, context=translation_context
        ),
        month_abbr=util.translate(
            util.month_msgid(dt.month(), "a"), domain=dom, context=translation_context
        ),
        week=dt.week(),
        wkday=dt.dow(),
        wkday_name=util.translate(
            util.day_msgid(dt.dow()), domain=dom, context=translation_context
        ),
        wkday_abbr=util.translate(
            util.day_msgid(dt.dow(), "s"), domain=dom, context=translation_context
        ),
        day=dt.day(),
        day2=zero_pad(dt.day()),
        hour=dt.hour(),
        hour2=zero_pad(dt.hour()),
        minute=dt.minute(),
        minute2=zero_pad(dt.minute()),
        second=dt.second(),
        second2=zero_pad(dt.second()),
    )
    return date_dict


def default_start(context=None):
    """Return the default start as python datetime for prefilling forms.

    :returns: Default start datetime.
    :rtype: Python datetime

    """
    now = localized_now(context=context)
    return now.replace(minute=0, second=0, microsecond=0)


def default_end(context=None):
    """Return the default end as python datetime for prefilling forms.

    :returns: Default end datetime.
    :rtype: Python datetime

    """
    return default_start(context=context) + timedelta(hours=DEFAULT_END_DELTA)


# General utils
# TODO: Better fits to CMFPlone. (Taken from CMFPlone's new syndication tool)


class AnnotationAdapter:
    """Abstract Base Class for an annotation storage.

    If the annotation wasn't set, it won't be created until the first attempt
    to set a property on this adapter.
    So, the context doesn't get polluted with annotations by accident.

    """

    ANNOTATION_KEY = None

    def __init__(self, context):
        self.context = context
        annotations = IAnnotations(context)
        self._data = annotations.get(self.ANNOTATION_KEY, None)

    def __setattr__(self, name, value):
        if name in ("context", "_data", "ANNOTATION_KEY"):
            self.__dict__[name] = value
        else:
            if self._data is None:
                self._data = PersistentDict()
                annotations = IAnnotations(self.context)
                annotations[self.ANNOTATION_KEY] = self._data
            self._data[name] = value

    def __getattr__(self, name):
        return self._data.get(name, None) if self._data else None


def find_context(context, viewname=None, iface=None, as_url=False, append_view=True):
    """Find the next context with a given view name or interface, up in the
    content tree, starting from the given context. This might not be the
    IPloneSiteRoot, but another subsite.

    :param context: The context to start the search from.
    :param viewname: (optional) The name of a view which a context should have
                     configured as defaultView.
    :param iface: (optional) The interface, the context to search for should
                  implement.
    :param as_url: (optional) Return the URL of the context found.
    :param append_view: (optional) In case of a given viewname and called with
                        as_url, append the viewname to the url, if the context
                        hasn't configured it as defaultView. Otherwise ignore
                        this parameter.
    :returns: A context with the given view name, interface or ISite root.
    """
    context = aq_inner(context)
    ret = None
    if (
        viewname
        and context.defaultView() == viewname
        or iface
        and iface.providedBy(context)
        or IPloneSiteRoot.providedBy(context)
    ):
        # Search for viewname or interface but stop at IPloneSiteRoot
        ret = context
    else:
        ret = find_context(
            aq_parent(context),
            viewname=viewname,
            iface=iface,
            as_url=False,
            append_view=False,
        )
    if as_url:
        url = ret.absolute_url()
        if viewname and append_view and ret.defaultView() != viewname:
            url = f"{url}/{viewname}"
        return url
    return ret


def find_site(context, as_url=False):
    return find_context(context, iface=ISite, as_url=as_url)


def find_ploneroot(context, as_url=False):
    return find_context(context, iface=IPloneSiteRoot, as_url=as_url)


def find_navroot(context, as_url=False):
    return find_context(context, iface=INavigationRoot, as_url=as_url)


def find_event_listing(context, as_url=False):
    return find_context(
        context, viewname="event_listing", iface=ISite, as_url=as_url, append_view=True
    )


# Workaround for buggy strftime with timezone handling in DateTime.
# See: https://github.com/plone/plone.app.event/pull/47
# TODO: should land in CMFPlone or fixed in DateTime.
_strftime = lambda v, fmt: pydt(v).strftime(fmt)  # noqa: E731


class PatchedDateTime(DateTime):
    def strftime(self, fmt):
        return _strftime(self, fmt)


def ulocalized_time(time, *args, **kwargs):
    """Corrects for DateTime bugs doing wrong thing with timezones"""
    wrapped_time = PatchedDateTime(time)
    return orig_ulocalized_time(wrapped_time, *args, **kwargs)



# setup dates
from datetime import datetime
from datetime import timedelta

import locale


# important for MONTHNAME
locale.setlocale(locale.LC_ALL, "en_US")

NOW = datetime.now() + timedelta(days=1)
EVENT_START_YEAR = NOW.strftime("%Y")
EVENT_START_MONTH = str(NOW.month - 1)
EVENT_START_MONTHNAME = NOW.strftime("%B")
EVENT_START_DAY = NOW.strftime("%d")
EVENT_START_HOUR = str(int(NOW.strftime("%I")))

END = NOW + timedelta(hours=1)
EVENT_END_YEAR = END.strftime("%Y")
EVENT_END_MONTH = str(END.month - 1)
EVENT_END_MONTHNAME = END.strftime("%B")
EVENT_END_DAY = END.strftime("%d")
EVENT_END_HOUR = str(int(END.strftime("%I")))



from AccessControl.Permission import addPermission
from zope.i18nmessageid import MessageFactory


packageName = __name__
_ = MessageFactory("plone")

# BBB Permissions
PORTAL_ADD_PERMISSION = "Add portal events"  # CMFCalendar/ATCT permissions

addPermission(
    PORTAL_ADD_PERMISSION,
    (
        "Manager",
        "Site Administrator",
        "Owner",
    ),
)



from plone.base.interfaces import INonInstallable
from Products.CMFCore.utils import getToolByName
from Products.ZCatalog.Catalog import CatalogError
from zope.interface import implementer

import logging


logger = logging.getLogger(__name__)


@implementer(INonInstallable)
class HiddenProfiles:
    def getNonInstallableProfiles(self):
        """Prevents profiles, which should not be user-installable from showing
        up in the profile list when creating a Plone site.

        plone.app.event:testing .. Testing profile, which provides an
        example type.
        """
        return [
            "plone.app.event:testing",
        ]


def setup_catalog(context):
    """Setup plone.app.event's indices in the catalog.

    Doing it here instead of in profiles/default/catalog.xml means we
    do not need to reindex those indexes after every reinstall.

    See these discussions for more info about index clearing with catalog.xml:
        http://plone.293351.n2.nabble.com/How-to-import-catalog-xml-without-
        emptying-the-indexes-td2302709.html
        https://mail.zope.org/pipermail/zope-cmf/2007-March/025664.html
    """
    if context.readDataFile("plone.app.event-default.txt") is None:
        return
    portal = context.getSite()
    catalog = getToolByName(portal, "portal_catalog")
    date_idxs = ["start", "end"]
    field_idxs = ["sync_uid"]
    idxs = date_idxs + field_idxs

    class extra:
        recurdef = "recurrence"
        until = ""

    _catalog = catalog._catalog
    for name in idxs:
        if name in catalog.indexes():
            if _catalog.getIndex(name).meta_type == "DateIndex":
                # delete old standard DateIndex
                catalog.delIndex(name)
                logger.info("Old catalog DateIndex %s deleted." % name)
        if name not in catalog.indexes():
            if name in date_idxs:
                # create new DateRecurringIndex
                catalog.addIndex(name, "DateRecurringIndex", extra=extra())
                logger.info("Catalog DateRecurringIndex %s created." % name)
            elif name in field_idxs:
                catalog.addIndex(name, "FieldIndex")
                logger.info("Catalog FieldIndex %s created." % name)
        try:
            catalog.addColumn(name)
            logger.info("Catalog metadata column %s created." % name)
        except CatalogError:
            logger.info("Catalog metadata column %s already exists." % name)



from Acquisition import aq_parent
from plone.app.contenttypes.behaviors.leadimage import ILeadImage
from plone.app.layout.viewlets import ViewletBase


class LeadImageViewlet(ViewletBase):
    """plone.app.contenttypes LeadImageViewlet for Occurrence contexts, where
    the image might be defined on the parent object.
    """

    def update(self):
        parent = aq_parent(self.context)
        self.available = (
            ILeadImage.providedBy(parent) and True
            if getattr(parent, "image", False)
            else False
        )



from calendar import monthrange
from datetime import date
from datetime import timedelta
from plone.app.contenttypes.behaviors.collection import ISyndicatableCollection
from plone.app.event import _
from plone.app.event.base import _prepare_range
from plone.app.event.base import expand_events
from plone.app.event.base import get_events
from plone.app.event.base import guess_date_from
from plone.app.event.base import localized_now
from plone.app.event.base import RET_MODE_ACCESSORS
from plone.app.event.base import RET_MODE_OBJECTS
from plone.app.event.base import spell_date
from plone.app.event.base import start_end_from_mode
from plone.app.event.base import start_end_query
from plone.app.event.ical.exporter import construct_icalendar
from plone.app.querystring import queryparser
from plone.base.batch import Batch
from plone.base.defaultpage import get_default_page
from plone.memoize import view
from plone.uuid.interfaces import IUUID
from Products.Five.browser import BrowserView
from zope.component import getMultiAdapter
from zope.contentprovider.interfaces import IContentProvider


class EventListing(BrowserView):
    def __init__(self, context, request):
        super().__init__(context, request)

        self.now = now = localized_now(context)

        # Try to get the default page
        default = get_default_page(context)
        self.default_context = context[default] if default else context

        self.is_collection = False
        if ISyndicatableCollection:
            self.is_collection = ISyndicatableCollection.providedBy(
                self.default_context
            )

        # Request parameter
        req = self.request.form

        b_size = int(req.get("b_size", 0))
        if not b_size and self.is_collection:
            collection_behavior = ISyndicatableCollection(self.default_context)
            b_size = getattr(collection_behavior, "item_count", 0)
        self.b_size = b_size or 10
        self.b_start = int(req.get("b_start", 0))
        self.orphan = int(req.get("orphan", 1))
        self.mode = req.get("mode", None)
        self._date = req.get("date", None)
        self.tags = req.get("tags", None)
        self.searchable_text = req.get("SearchableText", None)
        self.path = req.get("path", None)

        day = int(req.get("day", 0)) or None
        month = int(req.get("month", 0)) or None
        year = int(req.get("year", 0)) or None

        if not self._date and day or month or year:
            self._date = date(
                year or now.year, month or now.month, day or now.day
            ).isoformat()

        if self.mode is None:
            self.mode = "day" if self._date else "future"

        self.uid = None  # Used to get all occurrences from a single event. Overrides all other settings  # noqa

    @property
    def show_filter(self):
        ret = True
        if self.is_collection:
            ctx = self.default_context
            query = queryparser.parseFormquery(ctx, ctx.query)
            if "start" in query or "end" in query:
                # Don't show the date filter, if a date is given in the
                # collection's query
                ret = False
        return ret

    @property
    def date(self):
        dt = None
        if self._date:
            try:
                dt = guess_date_from(self._date)
            except TypeError:
                pass
        return dt

    @property
    def _start_end(self):
        start, end = start_end_from_mode(self.mode, self.date, self.context)
        return start, end

    def _get_events(self, ret_mode=RET_MODE_ACCESSORS, expand=True):
        context = self.context
        kw = {}
        if self.uid:
            # In this case, restrict search for single event
            kw["UID"] = self.uid
        else:
            if self.path:
                kw["path"] = self.path
            else:
                # Search current and subsequent folders
                kw["path"] = "/".join(context.getPhysicalPath())

            if self.tags:
                kw["Subject"] = {"query": self.tags, "operator": "and"}

            if self.searchable_text:
                kw["SearchableText"] = self.searchable_text

        # kw['b_start'] = self.b_start
        # kw['b_size']  = self.b_size

        start, end = self._start_end

        sort = "start"
        sort_reverse = False
        if self.mode in ("past", "all"):
            sort_reverse = True
        return get_events(
            context,
            start=start,
            end=end,
            sort=sort,
            sort_reverse=sort_reverse,
            ret_mode=ret_mode,
            expand=expand,
            **kw,
        )

    @view.memoize
    def events(self, ret_mode=RET_MODE_ACCESSORS, expand=True, batch=True):
        res = []
        if self.is_collection:
            ctx = self.default_context
            # Whatever sorting is defined, we're overriding it.
            sort_on = "start"
            sort_order = None
            if self.mode in ("past", "all"):
                sort_order = "reverse"
            query = queryparser.parseFormquery(
                ctx, ctx.query, sort_on=sort_on, sort_order=sort_order
            )
            custom_query = self.request.get("contentFilter", {})
            if "start" not in query or "end" not in query:
                # ... else don't show the navigation bar
                start, end = self._start_end
                start, end = _prepare_range(ctx, start, end)
                custom_query.update(start_end_query(start, end))
            res = ctx.results(batch=False, brains=True, custom_query=custom_query)
            if expand:
                # get start and end values from the query to ensure limited
                # listing for occurrences
                start, end = self._expand_events_start_end(
                    query.get("start") or custom_query.get("start"),
                    query.get("end") or custom_query.get("end"),
                )
                res = expand_events(
                    res,
                    ret_mode,
                    start=start,
                    end=end,
                    sort=sort_on,
                    sort_reverse=True if sort_order else False,
                )
        else:
            res = self._get_events(ret_mode, expand=expand)
        if batch:
            b_start = self.b_start
            b_size = self.b_size
            res = Batch(res, size=b_size, start=b_start, orphan=self.orphan)
        return res

    @property
    def ical(self):
        # Get as objects.
        # Don't include occurrences to avoid having them along with their
        # original events and it's recurrence definition in icalendar exports.
        events = self.events(ret_mode=RET_MODE_OBJECTS, expand=False, batch=False)
        cal = construct_icalendar(self.context, events)
        name = "%s.ics" % self.context.getId()
        contents = cal.to_ical()
        self.request.response.setHeader("Content-Type", "text/calendar")
        self.request.response.setHeader(
            "Content-Disposition", 'attachment; filename="%s"' % name
        )
        self.request.response.setHeader("Content-Length", len(contents))
        self.request.response.write(contents)

    @property
    def ical_url(self):
        date = self.date
        mode = self.mode

        qstr = "&".join(
            [
                it
                for it in [
                    "mode=%s" % mode if mode else None,
                    "date=%s" % date if date else None,
                ]
                if it
            ]
        )
        qstr = "?%s" % qstr if qstr else ""
        return f"{self.context.absolute_url()}/@@event_listing_ical{qstr}"

    # COLLECTION daterange start/end determination
    def _expand_events_start_end(self, start, end):
        # make sane start and end values for expand_events from
        # Collection start/end criterions.
        # if end/min is given, it overrides start/min settings to make sure,
        # ongoing events are shown in the listing!
        # XXX: This actually fits most needs, but not all. Maybe someone
        # wants to come up with some edgecases!
        se = dict(start=None, end=None)

        def safe_dt(val):
            # convert value from DateTime to datetime
            if hasattr(val, "asdatetime"):
                return val.asdatetime()
            return val

        if start:
            q = start.get("query")
            r = start.get("range")
            if r == "min":
                se["start"] = q
            elif r == "max":
                se["end"] = q
            elif r in ("minmax", "min:max"):
                list(q).sort()
                se["start"] = q[0]
                se["end"] = q[1]
        if end:
            q = end.get("query")
            r = end.get("range")
            if r == "min":
                se["start"] = q
        return safe_dt(se["start"]), safe_dt(se["end"])

    def formatted_date(self, occ):
        provider = getMultiAdapter(
            (self.context, self.request, self), IContentProvider, name="formatted_date"
        )
        return provider(occ)

    def date_speller(self, date):
        return spell_date(date, self.context)

    @property
    def header_string(self):
        start, end = self._start_end
        start_dict = spell_date(start, self.context) if start else None
        end_dict = spell_date(end, self.context) if end else None

        mode = self.mode
        main_msgid = None
        sub_msgid = None
        if mode == "all":
            main_msgid = _("all_events", default="All events")

        elif mode == "past":
            main_msgid = _("past_events", default="Past events")

        elif mode == "future":
            main_msgid = _("future_events", default="Future events")

        elif mode == "now":
            main_msgid = _("todays_upcoming_events", default="Todays upcoming events")

        elif mode == "today":
            main_msgid = _("todays_events", default="Todays events")

        elif mode == "7days":
            main_msgid = _("7days_events", default="Events in next 7 days.")
            sub_msgid = _(
                "events_from_until",
                default="${from} until ${until}.",
                mapping={
                    "from": "%s, %s. %s %s"
                    % (
                        start_dict["wkday_name"],
                        start.day,
                        start_dict["month_name"],
                        start.year,
                    ),
                    "until": "%s, %s. %s %s"
                    % (
                        end_dict["wkday_name"],
                        end.day,
                        end_dict["month_name"],
                        end.year,
                    ),
                },
            )

        elif mode == "day":
            main_msgid = _(
                "events_on_day",
                default="Events on ${day}",
                mapping={
                    "day": "%s, %s. %s %s"
                    % (
                        start_dict["wkday_name"],
                        start.day,
                        start_dict["month_name"],
                        start.year,
                    ),
                },
            )

        elif mode == "week":
            main_msgid = _(
                "events_in_week",
                default="Events in week ${weeknumber}",
                mapping={"weeknumber": start.isocalendar()[1]},
            )
            sub_msgid = _(
                "events_from_until",
                default="${from} until ${until}.",
                mapping={
                    "from": "%s, %s. %s %s"
                    % (
                        start_dict["wkday_name"],
                        start.day,
                        start_dict["month_name"],
                        start.year,
                    ),
                    "until": "%s, %s. %s %s"
                    % (
                        end_dict["wkday_name"],
                        end.day,
                        end_dict["month_name"],
                        end.year,
                    ),
                },
            )

        elif mode == "month":
            main_msgid = _(
                "events_in_month",
                default="Events in ${month} ${year}",
                mapping={
                    "month": start_dict["month_name"],
                    "year": start.year,
                },
            )

        trans = self.context.translate
        return {
            "main": trans(main_msgid) if main_msgid else "",
            "sub": trans(sub_msgid) if sub_msgid else "",
        }

    # MODE URLs
    def _date_nav_url(self, mode, datestr=""):
        return "{}?mode={}{}".format(
            self.request.getURL(),
            mode,
            "&date=%s" % datestr if datestr else "",
        )

    @property
    def mode_all_url(self):
        return self._date_nav_url("all")

    @property
    def mode_future_url(self):
        return self._date_nav_url("future")

    @property
    def mode_past_url(self):
        return self._date_nav_url("past")

    @property
    def mode_day_url(self):
        now = self.date or self.now
        return self._date_nav_url("day", now.date().isoformat())

    @property
    def mode_week_url(self):
        now = self.date or self.now
        return self._date_nav_url("week", now.date().isoformat())

    @property
    def mode_month_url(self):
        now = self.date or self.now
        return self._date_nav_url("month", now.date().isoformat())

    # DAY NAV
    @property
    def next_day_url(self):
        now = self.date or self.now
        datestr = (now + timedelta(days=1)).date().isoformat()
        return self._date_nav_url("day", datestr)

    @property
    def today_url(self):
        return self._date_nav_url("day")

    @property
    def prev_day_url(self):
        now = self.date or self.now
        datestr = (now - timedelta(days=1)).date().isoformat()
        return self._date_nav_url("day", datestr)

    # WEEK NAV
    @property
    def next_week_url(self):
        now = self.date or self.now
        datestr = (now + timedelta(days=7)).date().isoformat()
        return self._date_nav_url("week", datestr)

    @property
    def this_week_url(self):
        return self._date_nav_url("week")

    @property
    def prev_week_url(self):
        now = self.date or self.now
        datestr = (now - timedelta(days=7)).date().isoformat()
        return self._date_nav_url("week", datestr)

    # MONTH NAV
    @property
    def next_month_url(self):
        now = self.date or self.now
        last_day = monthrange(now.year, now.month)[1]  # (wkday, days)
        datestr = (now.replace(day=last_day) + timedelta(days=1)).date().isoformat()
        return self._date_nav_url("month", datestr)

    @property
    def this_month_url(self):
        return self._date_nav_url("month")

    @property
    def prev_month_url(self):
        now = self.date or self.now
        datestr = (now.replace(day=1) - timedelta(days=1)).date().isoformat()
        return self._date_nav_url("month", datestr)


class EventListingIcal(EventListing):
    def __call__(self, *args, **kwargs):
        return self.ical


class EventEventListing(EventListing):
    """This is an EventListing for an individual event, to list all
    occurrences batched and navigatable with all the features, the EventListing
    offers.
    """

    def __init__(self, context, request):
        super().__init__(context, request)
        self.uid = IUUID(self.context)



from zope.interface import Interface
from zope.publisher.interfaces.browser import IDefaultBrowserLayer


ISO_DATE_FORMAT = "%Y-%m-%d"


class IBrowserLayer(IDefaultBrowserLayer):
    """plone.app.event specific browser layer."""


class IICalendarImportEnabled(Interface):
    """Marker interface for contexts, where icalendar import is enabled."""



from plone.app.event.base import find_navroot
from plone.app.event.base import find_ploneroot


def get_calendar_url(context, search_base):
    # search_base is always from the portal_root object. We won't include
    # the path from the portal root object, so we traverse to the calendar
    # object and call it's url then.
    calendar_url = None
    if search_base:
        portal = find_ploneroot(context)
        search_base = "/".join(search_base.split("/")[2:])
        calendar_url = portal.unrestrictedTraverse(
            search_base.lstrip("/")  # start relative, first slash is omitted
        ).absolute_url()
    else:
        site_url = find_navroot(context, as_url=True)
        calendar_url = "%s/event_listing" % site_url

    return calendar_url



from datetime import datetime
from plone.app.event import base
from plone.app.event.dx.traverser import OccurrenceTraverser as OccTravDX
from plone.app.event.ical.exporter import add_to_zones_map
from plone.app.event.ical.importer import ical_import
from plone.app.event.testing import make_fake_response
from plone.app.event.testing import PAEventDX_FUNCTIONAL_TESTING
from plone.app.event.testing import PAEventDX_INTEGRATION_TESTING
from plone.app.event.tests.base_setup import AbstractSampleDataEvents
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.event.interfaces import IEventAccessor
from zope.component import getMultiAdapter

import os
import pytz
import unittest


# TODO:
# * test all event properties
# * enforce correct order: EXDATE and RDATE directly after RRULE
# TODO ical import browser tests


class ICalendarExportTestDX(AbstractSampleDataEvents):
    layer = PAEventDX_INTEGRATION_TESTING

    def traverser(self, context, request):
        return OccTravDX(context, request)

    def checkOrder(self, text, *order):
        for item in order:
            position = text.find(item)
            if position == -1:
                # ;VALUE=DATE-TIME is the default so it is optional.
                # icalendar will remove it in a future release.
                # See https://github.com/collective/icalendar/pull/450
                position = text.find(item.replace(";VALUE=DATE-TIME", ""))
            self.assertTrue(
                position >= 0, 'menu item "%s" missing or out of order' % item
            )
            text = text[position:]

    def test_event_ical(self):
        headers, output, request = make_fake_response(self.request)
        view = getMultiAdapter((self.now_event, request), name="ics_view")
        view()
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers["Content-Type"], "text/calendar")
        self.assertTrue("Content-Length" in headers)
        output = [i.decode("utf8") for i in output]
        icalstr = "".join(output)
        self.checkOrder(
            icalstr,
            "BEGIN:VCALENDAR",
            "VERSION:2.0",
            "PRODID:-//Plone.org//NONSGML plone.app.event//EN",
            "X-WR-TIMEZONE:Europe/Vienna",
            "BEGIN:VEVENT",
            "SUMMARY:Now Event",
            "DTSTART;TZID=Europe/Vienna;VALUE=DATE-TIME:20130505T100000",
            "DTEND;TZID=Europe/Vienna;VALUE=DATE-TIME:20130505T110000",
            "DTSTAMP;VALUE=DATE-TIME:",
            "UID:",
            "RRULE:FREQ=DAILY;COUNT=3;INTERVAL=1",
            "RDATE;TZID=Europe/Vienna:20130509T000000",
            "EXDATE;TZID=Europe/Vienna:20130506T000000,20140404T000000",
            "CATEGORIES:plone,testing",
            "CONTACT:Auto Testdriver\\, +123456789\\, testdriver@plone.org\\, http://plone",  # noqa
            " .org",  # line longer than max length spec by icalendar
            "CREATED;VALUE=DATE-TIME:",
            "LAST-MODIFIED;VALUE=DATE-TIME:",
            "LOCATION:Vienna",
            "URL:http://nohost/plone/now",
            "END:VEVENT",
            "BEGIN:VTIMEZONE",
            "TZID:Europe/Vienna",
            "X-LIC-LOCATION:Europe/Vienna",
            "BEGIN:DAYLIGHT",
            "DTSTART;VALUE=DATE-TIME:20130331T030000",
            "TZNAME:CEST",
            "TZOFFSETFROM:+0100",
            "TZOFFSETTO:+0200",
            "END:DAYLIGHT",
            "END:VTIMEZONE",
            "END:VCALENDAR",
        )

    def test_event_occurrence_ical(self):
        """A event occurrence should not contain recurrence definitions from
        it's parent.
        """
        headers, output, request = make_fake_response(self.request)
        occ = self.traverser(self.now_event, request).publishTraverse(
            request, "2013-05-07"
        )
        view = getMultiAdapter((occ, request), name="ics_view")
        view()
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers["Content-Type"], "text/calendar")
        self.assertTrue("Content-Length" in headers)
        output = [i.decode("utf8") for i in output]
        icalstr = "".join(output)
        self.assertTrue("Now Event" in icalstr)
        self.assertTrue("RRULE" not in icalstr)

    def test_portal_ical(self):
        headers, output, request = make_fake_response(self.request)
        view = getMultiAdapter((self.portal, request), name="ics_view")
        view()
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers["Content-Type"], "text/calendar")
        self.assertTrue("Content-Length" in headers)
        output = [i.decode("utf8") for i in output]
        icalstr = "".join(output)

        # No occurrences in export. Otherwise count would be 8.
        self.assertEqual(icalstr.count("BEGIN:VEVENT"), 4)
        self.checkOrder(
            icalstr,
            "BEGIN:VCALENDAR",
            "VERSION:2.0",
            "PRODID:-//Plone.org//NONSGML plone.app.event//EN",
            "X-WR-TIMEZONE:Europe/Vienna",
            # whole_day event
            "BEGIN:VEVENT",
            "SUMMARY:Past Event",
            "DTSTART;VALUE=DATE:20130425",
            "DTEND;VALUE=DATE:20130426",
            "DTSTAMP;VALUE=DATE-TIME:",
            "UID:",
            "RRULE:FREQ=DAILY;COUNT=3",
            "CREATED;VALUE=DATE-TIME:",
            "LAST-MODIFIED;VALUE=DATE-TIME:",
            "LOCATION:Vienna",
            "URL:http://nohost/plone/past",
            "END:VEVENT",
            "BEGIN:VEVENT",
            "SUMMARY:Long Event",
            "DTSTART;TZID=Europe/Vienna;VALUE=DATE-TIME:20130425T100000",
            "DTEND;TZID=Europe/Vienna;VALUE=DATE-TIME:20130604T100000",
            "DTSTAMP;VALUE=DATE-TIME:",
            "UID:",
            "CREATED;VALUE=DATE-TIME:",
            "LAST-MODIFIED;VALUE=DATE-TIME:",
            "LOCATION:Schaftal",
            "URL:http://nohost/plone/sub/long",
            "END:VEVENT",
            "BEGIN:VEVENT",
            "SUMMARY:Now Event",
            "DTSTART;TZID=Europe/Vienna;VALUE=DATE-TIME:20130505T100000",
            "DTEND;TZID=Europe/Vienna;VALUE=DATE-TIME:20130505T110000",
            "DTSTAMP;VALUE=DATE-TIME:",
            "UID:",
            "RRULE:FREQ=DAILY;COUNT=3;INTERVAL=1",
            "RDATE;TZID=Europe/Vienna:20130509T000000",
            "EXDATE;TZID=Europe/Vienna:20130506T000000,20140404T000000",
            "CATEGORIES:plone,testing",
            "CONTACT:Auto Testdriver\\, +123456789\\, testdriver@plone.org\\, http://plone",  # noqa
            " .org",
            "CREATED;VALUE=DATE-TIME:",
            "LAST-MODIFIED;VALUE=DATE-TIME:",
            "LOCATION:Vienna",
            "URL:http://nohost/plone/now",
            "END:VEVENT",
            "BEGIN:VEVENT",
            "SUMMARY:Future Event",
            "DTSTART;TZID=Europe/Vienna;VALUE=DATE-TIME:20130515T100000",
            "DTEND;TZID=Europe/Vienna;VALUE=DATE-TIME:20130515T110000",
            "DTSTAMP;VALUE=DATE-TIME:",
            "UID:",
            "CREATED;VALUE=DATE-TIME:",
            "LAST-MODIFIED;VALUE=DATE-TIME:",
            "LOCATION:Graz",
            "URL:http://nohost/plone/future",
            "END:VEVENT",
            "BEGIN:VTIMEZONE",
            "TZID:Europe/Vienna",
            "X-LIC-LOCATION:Europe/Vienna",
            "BEGIN:DAYLIGHT",
            "DTSTART;VALUE=DATE-TIME:20130331T030000",
            "TZNAME:CEST",
            "TZOFFSETFROM:+0100",
            "TZOFFSETTO:+0200",
            "END:DAYLIGHT",
            "END:VTIMEZONE",
            "END:VCALENDAR",
        )

    def test_event_listing_ical_portal(self):
        """Test event_listing ical export. It should contain all events from
        the listing, except Occurrences. For occurrences, their original events
        are exported.
        """
        headers, output, request = make_fake_response(self.request)
        view = getMultiAdapter((self.portal, request), name="event_listing_ical")
        view.mode = "all"
        view()
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers["Content-Type"], "text/calendar")
        self.assertTrue("Content-Length" in headers)
        output = [i.decode("utf8") for i in output]
        icalstr = "".join(output)
        # No occurrences in export. Otherwise count would be 8.
        self.assertEqual(icalstr.count("BEGIN:VEVENT"), 4)

    def test_event_listing_ical_portal__specific_date(self):
        """Test event_listing ical export for a specific date. The date is when
        a occurrence happens. It shouldn't contain the occurrence but the
        original event and the long lasting event.
        """
        headers, output, request = make_fake_response(self.request)
        view = getMultiAdapter((self.portal, request), name="event_listing_ical")
        view.mode = "day"
        view._date = "2013-04-27"
        view()
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers["Content-Type"], "text/calendar")
        self.assertTrue("Content-Length" in headers)
        output = [i.decode("utf8") for i in output]
        icalstr = "".join(output)
        self.assertEqual(icalstr.count("BEGIN:VEVENT"), 2)
        self.assertTrue("Past Event" in icalstr)
        self.assertTrue("Long Event" in icalstr)

    def test_collection_ical(self):
        """Test basic icalendar export from Collections."""
        headers, output, request = make_fake_response(self.request)
        view = getMultiAdapter((self.portal.collection, request), name="ics_view")
        view()
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers["Content-Type"], "text/calendar")
        self.assertTrue("Content-Length" in headers)
        output = [i.decode("utf8") for i in output]
        icalstr = "".join(output)
        self.assertEqual(icalstr.count("BEGIN:VEVENT"), 4)

    def test_collection_all_ical(self):
        """Test basic icalendar export from Collections, which returns not only
        events.
        """
        headers, output, request = make_fake_response(self.request)
        self.portal.collection.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.selection.any",
                "v": ["Event", "plone.app.event.dx.event", "Page"],
            },
        ]
        view = getMultiAdapter((self.portal.collection, request), name="ics_view")
        view()
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers["Content-Type"], "text/calendar")
        self.assertTrue("Content-Length" in headers)
        output = [i.decode("utf8") for i in output]
        icalstr = "".join(output)
        self.assertEqual(icalstr.count("BEGIN:VEVENT"), 4)


class TestIcalImportDX(unittest.TestCase):
    layer = PAEventDX_FUNCTIONAL_TESTING
    event_type = "plone.app.event.dx.event"

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

    def test_import_from_ics(self):
        # Ical import unit test.
        self.portal.invokeFactory("Folder", "impfolder1")
        impfolder = self.portal.impfolder1

        directory = os.path.dirname(__file__)
        with open(os.path.join(directory, "icaltest.ics"), "rb") as icsfile:
            icsdata = icsfile.read()
        res = ical_import(impfolder, icsdata, self.event_type)

        self.assertEqual(res["count"], 5)
        self.assertEqual(len(impfolder.contentIds()), 5)

        at = pytz.timezone("Europe/Vienna")
        utc = pytz.utc

        # Use pydt to normalize for DST times.

        # TODO: test for attendees. see note in
        # plone.app.event.ical.importer.ical_import
        e1 = IEventAccessor(impfolder.e1)
        self.assertEqual(e1.start, at.localize(datetime(2013, 7, 19, 12, 0)))
        self.assertEqual(e1.end, at.localize(datetime(2013, 7, 20, 13, 0)))
        self.assertEqual(e1.description, "A basic event with many properties.")
        self.assertEqual(e1.whole_day, False)
        self.assertEqual(e1.open_end, False)
        self.assertEqual(
            e1.sync_uid,
            "48f1a7ad64e847568d860cd092344970",
        )

        e2 = IEventAccessor(impfolder.e2)
        self.assertEqual(e2.start, utc.localize(datetime(1996, 4, 1, 1, 0)))
        self.assertEqual(e2.end, utc.localize(datetime(1996, 4, 1, 2, 0)))
        self.assertEqual(
            e2.recurrence,
            "RRULE:FREQ=DAILY;COUNT=100\nEXDATE:19960402T010000Z,"
            "19960403T010000Z,19960404T010000Z",
        )

        e3 = IEventAccessor(impfolder.e3)
        self.assertEqual(e3.start, at.localize(datetime(2012, 3, 27, 10, 0)))
        self.assertEqual(e3.end, at.localize(datetime(2012, 3, 27, 18, 0)))
        self.assertEqual(
            e3.recurrence,
            "RRULE:FREQ=WEEKLY;UNTIL=20120703T080000Z;BYDAY=TU\n"
            "EXDATE:20120529T100000,20120403T100000,20120410T100000,"
            "20120501T100000,20120417T100000",
        )

        e4 = IEventAccessor(impfolder.e4)
        self.assertEqual(e4.start, utc.localize(datetime(2013, 4, 4, 0, 0)))
        self.assertEqual(e4.end, utc.localize(datetime(2013, 4, 4, 23, 59, 59)))
        self.assertEqual(e4.whole_day, True)
        self.assertEqual(e4.open_end, False)

        e5 = IEventAccessor(impfolder.e5)
        self.assertEqual(e5.start, utc.localize(datetime(2013, 4, 2, 12, 0)))
        self.assertEqual(e5.end, utc.localize(datetime(2013, 4, 2, 23, 59, 59)))
        self.assertEqual(e5.whole_day, False)
        self.assertEqual(e5.open_end, True)

    def test_import_from_ics__no_sync(self):
        """SYNC_NONE and importing the same file again should create new event
        objects and give them each a new sync_uid.
        """
        self.portal.invokeFactory("Folder", "impfolder2")
        impfolder = self.portal.impfolder2

        directory = os.path.dirname(__file__)
        with open(os.path.join(directory, "icaltest.ics"), "rb") as icsfile:
            icsdata = icsfile.read()

        res = ical_import(impfolder, icsdata, self.event_type)

        self.assertEqual(res["count"], 5)

        e11 = impfolder["e1"]
        suid1 = IEventAccessor(e11).sync_uid

        res = ical_import(
            impfolder, icsdata, self.event_type, sync_strategy=base.SYNC_NONE
        )
        self.assertEqual(res["count"], 5)

        e12 = impfolder["e1-1"]
        suid2 = IEventAccessor(e12).sync_uid

        self.assertEqual(len(impfolder.contentIds()), 10)
        self.assertNotEqual(suid1, suid2)

    def test_import_from_ics__sync_keep_mine(self):
        """SYNC_KEEP_MINE and importing the same file again should do nothing."""
        self.portal.invokeFactory("Folder", "impfolder3")
        impfolder = self.portal.impfolder3

        directory = os.path.dirname(__file__)
        with open(os.path.join(directory, "icaltest.ics"), "rb") as icsfile:
            icsdata = icsfile.read()

        res = ical_import(impfolder, icsdata, self.event_type)
        self.assertEqual(res["count"], 5)

        e1a = IEventAccessor(impfolder.e1)
        mod1 = e1a.last_modified
        suid1 = e1a.sync_uid

        res = ical_import(
            impfolder, icsdata, self.event_type, sync_strategy=base.SYNC_KEEP_MINE
        )
        self.assertEqual(res["count"], 0)
        e1a = IEventAccessor(impfolder.e1)
        mod2 = e1a.last_modified
        suid2 = e1a.sync_uid

        self.assertEqual(len(impfolder.contentIds()), 5)

        self.assertEqual(mod1, mod2)
        self.assertEqual(suid1, suid2)

    def test_import_from_ics__sync_drop_older(self):
        """SYNC_KEEP_NEWER and importing the same file again should update only
        newer and on equal modified date but drop the change when it is older.
        """
        self.portal.invokeFactory("Folder", "impfolder4")
        impfolder = self.portal.impfolder4

        directory = os.path.dirname(__file__)
        with open(os.path.join(directory, "icaltest.ics"), "rb") as icsfile:
            icsdata1 = icsfile.read()

        with open(os.path.join(directory, "icaltest2.ics"), "rb") as icsfile:
            icsdata2 = icsfile.read()

        res = ical_import(impfolder, icsdata1, self.event_type)
        self.assertEqual(res["count"], 5)

        e1a = IEventAccessor(impfolder.e1)
        mod1 = e1a.last_modified
        suid1 = e1a.sync_uid
        title1 = e1a.title
        desc1 = e1a.description
        start1 = e1a.start
        end1 = e1a.end

        res = ical_import(
            impfolder, icsdata2, self.event_type, sync_strategy=base.SYNC_KEEP_NEWER
        )
        self.assertEqual(res["count"], 4)
        e1a = IEventAccessor(impfolder.e1)
        mod2 = e1a.last_modified
        suid2 = e1a.sync_uid
        title2 = e1a.title
        desc2 = e1a.description
        start2 = e1a.start
        end2 = e1a.end

        self.assertEqual(len(impfolder.contentIds()), 5)

        self.assertTrue(mod1 < mod2)
        self.assertEqual(suid1, suid2)
        self.assertNotEqual(title1, title2)
        self.assertNotEqual(desc1, desc2)
        self.assertTrue(start1 < start2)
        self.assertTrue(end1 < end2)

    def test_import_from_ics__sync_keep_theirs(self):
        """SYNC_KEEP_THEIRS and importing the same file again should update
        all.
        """
        self.portal.invokeFactory("Folder", "impfolder5")
        impfolder = self.portal.impfolder5

        directory = os.path.dirname(__file__)

        with open(os.path.join(directory, "icaltest.ics"), "rb") as icsfile:
            icsdata1 = icsfile.read()

        with open(os.path.join(directory, "icaltest2.ics"), "rb") as icsfile:
            icsdata2 = icsfile.read()

        res = ical_import(impfolder, icsdata1, self.event_type)
        self.assertEqual(res["count"], 5)

        e1a = IEventAccessor(impfolder.e1)
        mod11 = e1a.last_modified
        suid11 = e1a.sync_uid
        title11 = e1a.title
        desc11 = e1a.description
        start11 = e1a.start
        end11 = e1a.end

        e2a = IEventAccessor(impfolder.e2)
        suid21 = e2a.sync_uid
        title21 = e2a.title
        desc21 = e2a.description
        start21 = e2a.start
        end21 = e2a.end

        res = ical_import(
            impfolder, icsdata2, self.event_type, sync_strategy=base.SYNC_KEEP_THEIRS
        )
        self.assertEqual(res["count"], 5)

        e1a = IEventAccessor(impfolder.e1)
        mod12 = e1a.last_modified
        suid12 = e1a.sync_uid
        title12 = e1a.title
        desc12 = e1a.description
        start12 = e1a.start
        end12 = e1a.end

        e2a = IEventAccessor(impfolder.e2)
        suid22 = e2a.sync_uid
        title22 = e2a.title
        desc22 = e2a.description
        start22 = e2a.start
        end22 = e2a.end

        self.assertEqual(len(impfolder.contentIds()), 5)

        self.assertTrue(mod11 < mod12)
        self.assertEqual(suid11, suid12)
        self.assertNotEqual(title11, title12)
        self.assertNotEqual(desc11, desc12)
        self.assertTrue(start11 < start12)
        self.assertTrue(end11 < end12)

        self.assertEqual(suid21, suid22)
        self.assertNotEqual(title21, title22)
        self.assertNotEqual(desc21, desc22)
        self.assertTrue(start21 < start22)
        self.assertTrue(end21 < end22)

    def test_add_to_zones_map(self):
        tzmap = add_to_zones_map({}, "Africa/Abidjan", datetime(2013, 1, 1, 0, 0))
        self.assertTrue("Africa/Abidjan" in tzmap)



from plone.app.event.dx.traverser import OccurrenceTraverser as OccTravDX
from plone.app.event.testing import PAEventDX_FUNCTIONAL_TESTING
from plone.app.event.tests.base_setup import AbstractSampleDataEvents
from plone.app.event.tests.base_setup import patched_now as PN
from unittest import mock


class FunctionalTestEventViewDX(AbstractSampleDataEvents):
    layer = PAEventDX_FUNCTIONAL_TESTING

    @property
    def traverser(self):
        return OccTravDX(self.now_event, self.request)

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_event_view__non_recurring(self):
        view = self.portal.future.restrictedTraverse("@@event_view")

        output = view()

        self.assertTrue("Future Event" in output)
        self.assertTrue("2013-05-15" in output)
        self.assertTrue("Europe/Vienna" in output)
        self.assertTrue("UTC200" in output)
        self.assertTrue("Graz" in output)
        self.assertTrue("All dates" not in output)

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_event_view__recurring(self):
        view = self.portal.now.restrictedTraverse("@@event_view")

        output = view()

        self.assertTrue("Now Event" in output)
        self.assertTrue("2013-05-05" in output)
        self.assertTrue("All dates" in output)
        self.assertTrue("2013-05-07" in output)
        self.assertTrue("2013-05-09" in output)
        self.assertTrue("http://plone.org" in output)

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_event_view__recurring_occurrence(self):
        occ = self.traverser.publishTraverse(self.request, "2013-05-07")
        view = occ.restrictedTraverse("@@event_view")

        output = view()

        self.assertTrue("Now Event" in output)
        self.assertTrue(
            "2013-05-05" not in output
        )  # Lists only upcoming relative to occurrence's date  # noqa
        self.assertTrue("2013-05-07" in output)
        self.assertTrue("2013-05-09" in output)
        self.assertTrue("http://plone.org" in output)



from plone.app.event.dx.traverser import OccurrenceTraverser as OccTravDX
from plone.app.event.testing import PAEventDX_INTEGRATION_TESTING
from plone.app.event.tests.base_setup import AbstractSampleDataEvents
from plone.app.event.tests.base_setup import patched_now as PN
from unittest import mock


class TestEventSummaryDX(AbstractSampleDataEvents):
    layer = PAEventDX_INTEGRATION_TESTING

    @property
    def traverser(self):
        return OccTravDX(self.now_event, self.request)

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_event_summary__non_recurring(self):
        """Test if some specific content is included here."""
        view = self.portal.future.restrictedTraverse("@@event_summary")

        output = view()

        self.assertTrue("2013-05-15" in output)
        self.assertTrue("Europe/Vienna" in output)
        self.assertTrue("UTC200" in output)
        self.assertTrue("Graz" in output)
        self.assertTrue("All dates" not in output)

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_event_summary__recurring(self):
        """Test if some specific content is included here."""
        view = self.portal.now.restrictedTraverse("@@event_summary")

        self.assertEqual(view.is_occurrence, False)
        self.assertEqual(len(view.next_occurrences), 3)
        self.assertEqual(view.num_more_occurrences, 0)

        output = view()

        # self.assertTrue('Now Event' not in output)  # Title not shown by def.
        self.assertTrue("2013-05-05" in output)
        self.assertTrue("All dates" in output)
        self.assertTrue("2013-05-07" in output)
        self.assertTrue("2013-05-09" in output)
        self.assertTrue("http://plone.org" in output)

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_event_summary__recurring_occurrence(self):
        """Test if some specific content is included here."""
        occ = self.traverser.publishTraverse(self.request, "2013-05-07")
        view = occ.restrictedTraverse("@@event_summary")

        self.assertEqual(view.is_occurrence, True)
        # Lists only upcoming relative to occurrence's date
        self.assertEqual(len(view.next_occurrences), 2)
        # The num_more_occurrences link is shown as soon not all occurrences
        # are listed. Here. it's referencing the already passed occurrence.
        self.assertEqual(view.num_more_occurrences, 1)

        output = view()

        self.assertTrue("Now Event" not in output)  # Title not shown by def.
        self.assertTrue("All dates" in output)
        self.assertTrue(
            "2013-05-05" not in output
        )  # Lists only upcoming relative to occurrence's date  # noqa
        self.assertTrue("2013-05-07" in output)
        self.assertTrue("2013-05-09" in output)
        self.assertTrue("http://plone.org" in output)

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_event_summary__recurring_last_occurrence(self):
        """Test if some specific content is included here.

        On the last occurrence, no other occurrences should be listed but only
        the link to the event_listing view, where all occurrences of the event
        are shown.
        """
        occ = self.traverser.publishTraverse(self.request, "2013-05-09")
        view = occ.restrictedTraverse("@@event_summary")

        self.assertEqual(view.is_occurrence, True)
        # Lists only upcoming relative to occurrence's date. Here no other,
        # than the current (which is hidden from the occurrences list via the
        # template).
        self.assertEqual(len(view.next_occurrences), 1)
        # The num_more_occurrences link is shown as soon not all occurrences
        # are listed. Here. it's referencing the already passed occurrence.
        self.assertEqual(view.num_more_occurrences, 2)

        output = view()

        self.assertTrue("Now Event" not in output)  # Title not shown by def.
        self.assertTrue("All dates" in output)
        self.assertTrue("2013-05-05" not in output)
        self.assertTrue("2013-05-07" not in output)
        self.assertTrue("2013-05-09" in output)
        self.assertTrue("http://plone.org" in output)

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_event_summary__recurring_excludes(self):
        """Test if some specific content is excluded here."""
        view = self.portal.now.restrictedTraverse("@@event_summary")
        ex = view.excludes
        view.excludes = ["occurrences", "event_url"]

        output = view()

        self.assertTrue("Now Event" not in output)  # Title not shown by def.
        self.assertTrue("2013-05-05" in output)
        self.assertTrue("All dates" not in output)
        self.assertTrue("2013-05-07" not in output)
        self.assertTrue("2013-05-09" not in output)
        self.assertTrue("http://plone.org" not in output)

        # Restore default excludes
        view.excludes = ex



from plone.app.event.ical.importer import IcalendarImportSettingsFormView
from plone.app.event.interfaces import IICalendarImportEnabled
from plone.app.event.testing import PAEventDX_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.testing.zope import Browser

import transaction
import unittest


class TestICALImportSettings(unittest.TestCase):
    layer = PAEventDX_FUNCTIONAL_TESTING

    def setUp(self):
        app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.browser = Browser(app)
        self.browser.handleErrors = False
        self.browser.addHeader(
            "Authorization",
            "Basic %s:%s"
            % (
                SITE_OWNER_NAME,
                SITE_OWNER_PASSWORD,
            ),
        )

    def test_enable_ical_import(self):
        """Test that ical import can be enabled/disabled in the browser.
        Failed in Zope4: https://github.com/zopefoundation/Zope/issues/397
        """
        self.portal.invokeFactory("Folder", "f1")
        f1 = self.portal["f1"]
        self.assertFalse(IICalendarImportEnabled.providedBy(f1))

        # enable/disable with the view
        view = IcalendarImportSettingsFormView(f1, self.request)
        view.enable()
        self.assertTrue(IICalendarImportEnabled.providedBy(f1))
        view.disable()
        self.assertFalse(IICalendarImportEnabled.providedBy(f1))

        # enable/disable with traversal
        enable_method = f1.restrictedTraverse("ical_import_settings/enable")
        enable_method()
        self.assertTrue(IICalendarImportEnabled.providedBy(f1))
        disable_method = f1.restrictedTraverse("ical_import_settings/disable")
        disable_method()
        self.assertFalse(IICalendarImportEnabled.providedBy(f1))

        # enable/disable in the browser
        transaction.commit()
        f1_url = f1.absolute_url()
        self.assertFalse(IICalendarImportEnabled.providedBy(f1))
        self.browser.open(f1_url + "/ical_import_settings/enable")
        self.browser.getControl("Confirm action").click()
        self.assertTrue(IICalendarImportEnabled.providedBy(f1))
        self.browser.open(f1_url + "/ical_import_settings/disable")
        self.browser.getControl("Confirm action").click()
        self.assertFalse(IICalendarImportEnabled.providedBy(f1))

        # the form can be rendered
        self.browser.open(f1_url + "/ical_import_settings")
        self.assertIn(
            "URL to an external icalendar resource file", self.browser.contents
        )

    def test_constraint(self):
        self.portal.invokeFactory("Folder", "f1")
        f1 = self.portal["f1"]
        f1_url = f1.absolute_url()
        transaction.commit()

        # Enable ical import.
        self.browser.open(f1_url + "/ical_import_settings/enable")
        self.browser.getControl("Confirm action").click()

        # Set it to a file url.
        self.browser.open(f1_url + "/ical_import_settings")
        self.assertIn(
            "URL to an external icalendar resource file", self.browser.contents
        )
        self.browser.getControl(name="form.widgets.ical_url").value = (
            "file:///tmp/test.ical"
        )
        self.browser.getControl(name="form.buttons.save").click()
        self.assertIn("URLs with file: are not allowed.", self.browser.contents)



from plone.app.event.testing import PAEvent_INTEGRATION_TESTING
from Products.CMFCore.utils import getToolByName

import unittest


class TextDXIntegration(unittest.TestCase):
    layer = PAEvent_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.catalog = getToolByName(self.portal, "portal_catalog")

    def test_end_is_DateRecurringIndex(self):
        # end should be a DateRecurringIndex
        self.assertEqual(
            self.catalog.Indexes["end"].__class__.__name__, "DateRecurringIndex"
        )

    def test_start_is_DateRecurringIndex(self):
        # start should be a DateRecurringIndex
        self.assertEqual(
            self.catalog.Indexes["start"].__class__.__name__, "DateRecurringIndex"
        )



from datetime import datetime
from datetime import timedelta
from DateTime import DateTime
from OFS.SimpleItem import SimpleItem
from plone.app.event import base
from plone.app.event.base import get_events
from plone.app.event.base import localized_now
from plone.app.event.dx.behaviors import default_end
from plone.app.event.dx.behaviors import default_start
from plone.app.event.dx.behaviors import IEventBasic
from plone.app.event.dx.behaviors import IEventRecurrence
from plone.app.event.dx.behaviors import StartBeforeEnd
from plone.app.event.dx.interfaces import IDXEvent
from plone.app.event.dx.interfaces import IDXEventRecurrence
from plone.app.event.testing import PAEventDX_FUNCTIONAL_TESTING
from plone.app.event.testing import PAEventDX_INTEGRATION_TESTING
from plone.app.event.testing import set_browserlayer
from plone.app.event.testing import set_env_timezone
from plone.app.event.tests.base_setup import patched_now
from plone.app.event.upgrades.upgrades import upgrade_attribute_storage
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.app.textfield.value import RichTextValue
from plone.dexterity.utils import createContentInContainer
from plone.event import utils
from plone.event.interfaces import IEvent
from plone.event.interfaces import IEventAccessor
from plone.event.interfaces import IOccurrence
from plone.event.interfaces import IRecurrenceSupport
from plone.testing.zope import Browser
from plone.uuid.interfaces import IUUID
from unittest import mock
from zope.annotation.interfaces import IAnnotations

import pytz
import unittest
import zope.interface


TEST_TIMEZONE = "Europe/Vienna"


class MockEvent(SimpleItem):
    """Mock event"""


class TestDXAddEdit(unittest.TestCase):
    layer = PAEventDX_FUNCTIONAL_TESTING

    def setUp(self):
        app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.browser = Browser(app)
        self.browser.handleErrors = False
        self.browser.addHeader(
            "Authorization",
            "Basic %s:%s"
            % (
                SITE_OWNER_NAME,
                SITE_OWNER_PASSWORD,
            ),
        )

    @mock.patch("plone.app.event.base.localized_now", new=patched_now)
    def test_defaults(self):
        """Test, if defaults are set correctly.

        Would have loved to do it like follows, but that doesn't set defaults:

        TRY 1:
        self.portal.invokeFactory(
            'plone.app.event.dx.event',
            'event1',
            title=u"event1",
        )

        TRY 2:
        from plone.dexterity.utils import createContentInContainer
        e1 = createContentInContainer(
            self.portal,
            'plone.app.event.dx.event',
            u'event1'
        )

        TRY 3:
        from plone.dexterity.browser.add import DefaultAddForm
        form = DefaultAddForm(self.portal, self.request)
        form.portal_type = 'plone.app.event.dx.event'
        e1 = form.create(data={'title': 'event1'})
        form.add(e1)
        """

        self.browser.open(
            "{}/{}".format(
                self.portal.absolute_url(), "++add++plone.app.event.dx.event"
            )
        )

        self.browser.getControl(name="form.widgets.IDublinCore.title").value = (
            "TestEvent"
        )

        # TODO: these values are simply not set in the pat-pickadate pattern.
        self.browser.getControl(name="form.widgets.IEventBasic.start").value = (
            "2014-03-30"
        )
        self.browser.getControl(name="form.widgets.IEventBasic.end").value = (
            "2014-03-31"
        )

        self.browser.getControl("Save").click()

        # CHECK VALUES
        #
        # TODO: fix all defaults
        event = self.portal["testevent"]
        # self.assertEqual(
        #     event.start,
        #     patched_now()
        # )
        # self.assertEqual(
        #     event.end,
        #     patched_now() + timedelta(hours=DEFAULT_END_DELTA)
        # )
        self.assertEqual(event.whole_day, False)
        self.assertEqual(event.open_end, False)
        # self.assertEqual(event.timezone, TEST_TIMEZONE)

    def test_edit_context(self):
        """Test if already added event can be edited directly on the context as
        intended.
        If should not fail with a timezone related error.
        """
        """
        self.portal.invokeFactory(
            'plone.app.event.dx.event',
            'testevent',
            title="Test Event",
            start=datetime(2014, 03, 29, 21, 53),
            end=datetime(2014, 03, 29, 22, 45),
            timezone=TEST_TIMEZONE
        )

        from plone.dexterity.browser.edit import DefaultEditForm
        # DOES NOT WORK...
        testevent = self.portal.testevent
        request = self.request
        request.form = {
            'form.widgets.IEventBasic.start': ('2014', '2', '2', '10', '10')
        }
        edit = DefaultEditForm(testevent, request)
        edit.update()

        save = edit.buttons['save']
        edit.handlers.getHandler(save)(edit, edit)
        """

        #
        # ADD
        #
        self.browser.open(self.portal.absolute_url())
        self.browser.getLink("plone.app.event.dx.event").click()
        self.browser.getControl(name="form.widgets.IDublinCore.title").value = (
            "TestEvent"
        )

        self.browser.getControl(name="form.widgets.IEventBasic.start").value = (
            "2014-03-30T03:51"
        )

        self.browser.getControl(name="form.widgets.IEventBasic.end").value = (
            "2014-03-30T04:51"
        )

        self.browser.getControl("Save").click()

        # CHECK VALUES
        #
        self.assertTrue(self.browser.url.endswith("testevent/view"))
        self.assertTrue("TestEvent" in self.browser.contents)
        self.assertTrue("2014-03-30" in self.browser.contents)

        #
        # EDIT
        #
        testevent = self.portal.testevent
        self.browser.open("%s/@@edit" % testevent.absolute_url())

        self.browser.getControl(name="form.widgets.IEventBasic.start").value = (
            "2014-03-31T03:51"
        )

        self.browser.getControl(name="form.widgets.IEventBasic.end").value = (
            "2014-03-31T04:51"
        )

        self.browser.getControl("Save").click()

        #
        # EDIT AGAIN
        #
        testevent = self.portal.testevent
        self.browser.open("%s/@@edit" % testevent.absolute_url())

        self.browser.getControl("Save").click()

        # CHECK DATES/TIMES, MUST NOT HAVE CHANGED
        #
        self.assertTrue("2014-03-31" in self.browser.contents)
        self.assertTrue("03:51" in self.browser.contents)
        self.assertTrue("04:51" in self.browser.contents)

        #
        # EDIT and set whole_day setting
        #
        testevent = self.portal.testevent
        self.browser.open("%s/@@edit" % testevent.absolute_url())

        self.browser.getControl(
            name="form.widgets.IEventBasic.whole_day:list"
        ).value = True

        self.browser.getControl("Save").click()

        # CHECK DATES/TIMES, IF THEY ADAPTED ACCORDING TO WHOLE DAY
        #
        self.assertTrue("2014-03-31" in self.browser.contents)
        self.assertTrue("0:00" in self.browser.contents)
        self.assertTrue("23:59" in self.browser.contents)


class TestEventAccessor(unittest.TestCase):
    layer = PAEventDX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        set_browserlayer(self.request)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

    def test_event_accessor(self):
        tz = pytz.timezone("Europe/Vienna")
        e1 = createContentInContainer(
            self.portal,
            "plone.app.event.dx.event",
            title="event1",
            start=tz.localize(datetime(2011, 11, 11, 11, 0)),
            end=tz.localize(datetime(2011, 11, 11, 12, 0)),
        )

        # setting attributes via the accessor
        acc = IEventAccessor(e1)
        new_end = tz.localize(datetime(2011, 11, 13, 10, 0))
        acc.end = new_end

        # context's end should be set to new_end
        self.assertEqual(e1.end, new_end)

        # accessor's and context datetime should be the same
        self.assertEqual(acc.end, e1.end)

    def test_event_accessor_whole_day__open_end(self):
        """Also ensures, that accessor method is called for getting start/end
        instead of a custom __getattr__ version.
        """
        at = pytz.timezone("Europe/Vienna")

        start = at.localize(datetime(2012, 10, 19, 0, 30))
        end = at.localize(datetime(2012, 10, 19, 1, 30))

        start_start = at.localize(datetime(2012, 10, 19, 0, 0, 0))
        end_end = at.localize(datetime(2012, 10, 19, 23, 59, 59))

        e1 = createContentInContainer(
            self.portal,
            "plone.app.event.dx.event",
            title="event1",
            start=start,
            end=end,
        )
        acc = IEventAccessor(e1)

        # check set
        self.assertEqual(e1.start, start)
        self.assertEqual(e1.end, end)

        # Setting open end
        e1.open_end = True
        self.assertEqual(e1.start, start)
        self.assertEqual(e1.end, end)
        self.assertEqual(acc.start, start)
        self.assertEqual(acc.end, end_end)

        # Setting whole day
        e1.whole_day = True
        self.assertEqual(e1.start, start)
        self.assertEqual(e1.end, end)
        self.assertEqual(acc.start, start_start)
        self.assertEqual(acc.end, end_end)

    def test_event_accessor__sync_uid(self):
        self.request.set("HTTP_HOST", "nohost")

        e1 = createContentInContainer(
            self.portal, "plone.app.event.dx.event", title="event1"
        )
        acc = IEventAccessor(e1)

        # setting no sync uid will automatically generate one
        self.assertTrue(acc.sync_uid, IUUID(e1) + "@nohost")
        # it's not stored on the object though
        self.assertEqual(e1.sync_uid, None)
        # but it's indexed
        result = self.portal.portal_catalog(sync_uid=IUUID(e1) + "@nohost")
        self.assertEqual(len(result), 1)

        # Setting the sync_uid
        acc.sync_uid = "okay"
        e1.reindexObject()
        self.assertEqual(acc.sync_uid, "okay")
        # Now, it's also stored on the object itself
        self.assertEqual(e1.sync_uid, "okay")
        # and indexed
        result = self.portal.portal_catalog(sync_uid="okay")
        self.assertEqual(len(result), 1)

    def test_event_accessor__start_end(self):
        e1 = createContentInContainer(
            self.portal, "plone.app.event.dx.event", title="event1"
        )

        dt = datetime(2161, 1, 1)  # United Federation of Planets
        DT = DateTime("2161/01/01 00:00:00 UTC")

        acc = IEventAccessor(e1)

        # Setting a timezone-naive datetime should convert it to UTC
        acc.start = dt
        self.assertEqual(acc.start, utils.utc(dt))
        self.assertEqual(e1.start, utils.utc(dt))
        # Setting a DateTime should convert it to datetime
        acc.start = DT
        self.assertEqual(acc.start, utils.utc(dt))
        self.assertEqual(e1.start, utils.utc(dt))

        # Same goes for acc.end
        # Setting a timezone-naive datetime should convert it to UTC
        acc.end = dt
        self.assertEqual(acc.end, utils.utc(dt))
        self.assertEqual(e1.end, utils.utc(dt))
        # Setting a DateTime should convert it to datetime
        acc.end = DT
        self.assertEqual(acc.end, utils.utc(dt))
        self.assertEqual(e1.end, utils.utc(dt))


class TestDXIntegration(unittest.TestCase):
    layer = PAEventDX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        set_browserlayer(self.request)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.tz = pytz.timezone(TEST_TIMEZONE)

    def test_start_defaults(self):
        data = MockEvent()
        data.context = MockEvent()
        default_value = default_start(data)
        now = localized_now().replace(minute=0, second=0, microsecond=0)
        delta = default_value - now
        self.assertEqual(0, delta.seconds)

    def test_end_default(self):
        data = MockEvent()
        data.context = MockEvent()
        default_value = default_end(data)
        delta = default_value - default_start(data)
        self.assertEqual(3600, delta.seconds)

    def test_start_end_dates_indexed(self):
        tz = pytz.timezone("Europe/Vienna")
        start = tz.localize(datetime(2011, 11, 11, 11, 0))
        end = tz.localize(datetime(2011, 11, 11, 12, 0))
        e1 = createContentInContainer(
            self.portal,
            "plone.app.event.dx.event",
            title="event1",
            start=start,
            end=end,
        )

        result = self.portal.portal_catalog(path="/".join(e1.getPhysicalPath()))
        self.assertEqual(1, len(result))

        # The start and end datetime's are indexed as Python datetimes
        self.assertEqual(result[0].start, start)
        self.assertEqual(result[0].end, end)

    def test_recurrence_indexing(self):
        tz = pytz.timezone("Europe/Vienna")
        e1 = createContentInContainer(
            self.portal,
            "plone.app.event.dx.event",
            title="event1",
            start=tz.localize(datetime(2011, 11, 11, 11, 0)),
            end=tz.localize(datetime(2011, 11, 11, 12, 0)),
        )

        # When editing via behaviors, the attributes should also be available
        # on the context itself.
        IEventRecurrence(e1).recurrence = "RRULE:FREQ=DAILY;COUNT=4"
        self.assertTrue(e1.recurrence == IEventRecurrence(e1).recurrence)

        e1.reindexObject()

        # get_events, no expanded occurrences
        result = get_events(self.portal)
        self.assertEqual(len(result), 1)

        # Get all the occurrences
        result = get_events(
            self.portal,
            start=tz.localize(datetime(2011, 11, 11, 11, 0)),
            ret_mode=base.RET_MODE_OBJECTS,
            expand=True,
        )
        self.assertEqual(len(result), 4)


class TestDXEventRecurrence(unittest.TestCase):
    layer = PAEventDX_INTEGRATION_TESTING

    def test_recurrence(self):
        tz = pytz.timezone("Europe/Vienna")
        duration = timedelta(days=4)
        mock = MockEvent()
        mock.start = tz.localize(datetime(2011, 11, 11, 11, 0))
        mock.end = mock.start + duration
        mock.recurrence = "RRULE:FREQ=DAILY;COUNT=4"
        zope.interface.alsoProvides(
            mock, IEvent, IEventBasic, IEventRecurrence, IDXEvent, IDXEventRecurrence
        )
        result = IRecurrenceSupport(mock).occurrences()
        result = list(result)  # cast generator to list

        self.assertEqual(4, len(result))

        # First occurrence is an IEvent object
        self.assertTrue(IEvent.providedBy(result[0]))

        # Subsequent ones are IOccurrence objects
        self.assertTrue(IOccurrence.providedBy(result[1]))


class TestDXEventUnittest(unittest.TestCase):
    """Unit test for Dexterity event behaviors."""

    def setUp(self):
        self.orig_tz = set_env_timezone(TEST_TIMEZONE)

    def tearDown(self):
        set_env_timezone(self.orig_tz)

    def test_validate_invariants_ok(self):
        mock = MockEvent()
        mock.start = datetime(2009, 1, 1)
        mock.end = datetime(2009, 1, 2)

        try:
            IEventBasic.validateInvariants(mock)
        except Exception:
            self.fail()

    def test_validate_invariants_fail(self):
        mock = MockEvent()
        mock.start = datetime(2009, 1, 2)
        mock.end = datetime(2009, 1, 1)
        mock.open_end = False

        try:
            IEventBasic.validateInvariants(mock)
            self.fail()
        except StartBeforeEnd:
            pass

    def test_validate_invariants_edge(self):
        mock = MockEvent()
        mock.start = datetime(2009, 1, 2)
        mock.end = datetime(2009, 1, 2)
        mock.open_end = False

        try:
            IEventBasic.validateInvariants(mock)
        except Exception:
            self.fail()

    def test_validate_invariants_openend(self):
        mock = MockEvent()
        mock.start = datetime(2009, 1, 2)
        mock.end = datetime(2009, 1, 1)
        mock.open_end = True

        try:
            IEventBasic.validateInvariants(mock)
        except Exception:
            self.fail()

    def test_validate_dont_validate_incomplete(self):
        """Don't validate validate_start_end invariant, if start or end are
        None.
        """
        mock = MockEvent()
        mock.open_end = False

        mock.start = datetime(2016, 5, 18)
        mock.end = None
        try:
            IEventBasic.validateInvariants(mock)
        except Exception:
            self.fail()

        mock.start = None
        mock.end = datetime(2016, 5, 18)
        try:
            IEventBasic.validateInvariants(mock)
        except Exception:
            self.fail()

        mock.start = None
        mock.end = None
        try:
            IEventBasic.validateInvariants(mock)
        except Exception:
            self.fail()


class TestDXAnnotationStorageUpdate(unittest.TestCase):
    """Unit tests for the Annotation Storage migration"""

    layer = PAEventDX_INTEGRATION_TESTING

    location = "Köln"
    attendees = ("Peter", "Søren", "Madeleine")
    contact_email = "person@email.com"
    contact_name = "Peter Parker"
    contact_phone = "555 123 456"
    event_url = "http://my.event.url"
    text = "<p>Cathedral Sprint in Köln</p>"

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        set_browserlayer(self.request)
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

    def test_migrate_fields(self):
        tz = pytz.timezone("Europe/Berlin")
        e1 = createContentInContainer(
            self.portal,
            "Event",
            title="event1",
            start=tz.localize(datetime(2011, 11, 11, 11, 0)),
            end=tz.localize(datetime(2011, 11, 11, 12, 0)),
        )

        # Fill the field values into the annotation storage
        ann = IAnnotations(e1)
        ann["plone.app.event.dx.behaviors.IEventLocation.location"] = self.location
        ann["plone.app.event.dx.behaviors.IEventAttendees.attendees"] = self.attendees
        ann["plone.app.event.dx.behaviors.IEventContact.contact_email"] = (
            self.contact_email
        )
        ann["plone.app.event.dx.behaviors.IEventContact.contact_name"] = (
            self.contact_name
        )
        ann["plone.app.event.dx.behaviors.IEventContact.contact_phone"] = (
            self.contact_phone
        )
        ann["plone.app.event.dx.behaviors.IEventContact.event_url"] = self.event_url
        ann["plone.app.event.dx.behaviors.IEventSummary.text"] = RichTextValue(
            raw=self.text
        )

        # All behavior-related fields are not set yet
        self.assertEqual(e1.location, None)
        self.assertEqual(e1.attendees, ())
        self.assertEqual(e1.contact_email, None)
        self.assertEqual(e1.contact_name, None)
        self.assertEqual(e1.contact_phone, None)
        self.assertEqual(e1.event_url, None)
        self.assertEqual(e1.text, None)

        # Run the upgrade step
        upgrade_attribute_storage(self.portal)

        # All behavior-related fields have been migrated
        self.assertEqual(e1.location, self.location)
        self.assertEqual(e1.attendees, self.attendees)
        self.assertEqual(e1.contact_email, self.contact_email)
        self.assertEqual(e1.contact_name, self.contact_name)
        self.assertEqual(e1.contact_phone, self.contact_phone)
        self.assertEqual(e1.event_url, self.event_url)
        self.assertEqual(e1.text.raw, self.text)

    def test_no_overwrite(self):
        tz = pytz.timezone("Europe/Berlin")
        e1 = createContentInContainer(
            self.portal,
            "Event",
            title="event1",
            start=tz.localize(datetime(2011, 11, 11, 11, 0)),
            end=tz.localize(datetime(2011, 11, 11, 12, 0)),
        )

        # Fill the field values into the annotation storage
        ann = IAnnotations(e1)
        ann["plone.app.event.dx.behaviors.IEventLocation.location"] = (
            self.location + "X"
        )
        ann["plone.app.event.dx.behaviors.IEventAttendees.attendees"] = (
            self.attendees + ("Paula",)
        )
        ann["plone.app.event.dx.behaviors.IEventContact.contact_email"] = (
            self.contact_email + "X"
        )
        ann["plone.app.event.dx.behaviors.IEventContact.contact_name"] = (
            self.contact_name + "X"
        )
        ann["plone.app.event.dx.behaviors.IEventContact.contact_phone"] = (
            self.contact_phone + "X"
        )
        ann["plone.app.event.dx.behaviors.IEventContact.event_url"] = (
            self.event_url + "X"
        )
        ann["plone.app.event.dx.behaviors.IEventSummary.text"] = RichTextValue(
            raw=self.text + "X"
        )

        # Add values into the fields in the new way
        e1.location = self.location
        e1.attendees = self.attendees
        e1.contact_email = self.contact_email
        e1.contact_phone = self.contact_phone
        e1.contact_name = self.contact_name
        e1.event_url = self.event_url
        e1.text = RichTextValue(raw=self.text)

        upgrade_attribute_storage(self.portal)

        # The already existing field values were not touched by the upgrade
        self.assertEqual(e1.location, self.location)
        self.assertEqual(e1.attendees, self.attendees)
        self.assertEqual(e1.contact_email, self.contact_email)
        self.assertEqual(e1.contact_phone, self.contact_phone)
        self.assertEqual(e1.contact_name, self.contact_name)
        self.assertEqual(e1.event_url, self.event_url)
        self.assertEqual(e1.text.raw, self.text)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)



from Acquisition import aq_inner
from datetime import datetime
from datetime import timedelta
from plone.app.contentlisting.interfaces import IContentListingObject
from plone.app.event.base import default_timezone
from plone.app.event.base import get_events
from plone.app.event.base import RET_MODE_BRAINS
from plone.event.interfaces import IEvent
from plone.event.interfaces import IEventAccessor
from plone.event.interfaces import IICalendar
from plone.event.interfaces import IICalendarEventComponent
from plone.event.interfaces import IOccurrence
from plone.event.utils import is_datetime
from plone.event.utils import tzdel
from plone.event.utils import utc
from Products.ZCatalog.interfaces import ICatalogBrain
from zope.interface import implementer
from zope.publisher.browser import BrowserView

import icalendar
import pytz


PRODID = "-//Plone.org//NONSGML plone.app.event//EN"
VERSION = "2.0"


def construct_icalendar(context, events):
    """Returns an icalendar.Calendar object.

    :param context: A content object, which is used for calendar details like
                    Title and Description. Usually a container, collection or
                    the event itself.

    :param events: The list of event objects, which are included in this
                   calendar.
    """
    cal = icalendar.Calendar()
    cal.add("prodid", PRODID)
    cal.add("version", VERSION)

    cal_tz = default_timezone(context)
    if cal_tz:
        cal.add("x-wr-timezone", cal_tz)

    tzmap = {}
    if not getattr(events, "__getitem__", False):
        events = [events]
    for event in events:
        if ICatalogBrain.providedBy(event) or IContentListingObject.providedBy(event):
            event = event.getObject()
        if not (IEvent.providedBy(event) or IOccurrence.providedBy(event)):
            # Must be an event.
            continue
        acc = IEventAccessor(event)
        tz = acc.timezone
        # TODO: the standard wants each recurrence to have a valid timezone
        # definition. sounds decent, but not realizable.
        if not acc.whole_day:  # whole day events are exported as dates without
            # timezone information
            if isinstance(tz, tuple):
                tz_start, tz_end = tz
            else:
                tz_start = tz_end = tz
            tzmap = add_to_zones_map(tzmap, tz_start, acc.start)
            tzmap = add_to_zones_map(tzmap, tz_end, acc.end)
        cal.add_component(IICalendarEventComponent(event).to_ical())

    for tzid, transitions in tzmap.items():
        cal_tz = icalendar.Timezone()
        cal_tz.add("tzid", tzid)
        cal_tz.add("x-lic-location", tzid)

        for transition, tzinfo in transitions.items():
            if tzinfo["dst"]:
                cal_tz_sub = icalendar.TimezoneDaylight()
            else:
                cal_tz_sub = icalendar.TimezoneStandard()

            cal_tz_sub.add("tzname", tzinfo["name"])
            cal_tz_sub.add("dtstart", transition)
            cal_tz_sub.add("tzoffsetfrom", tzinfo["tzoffsetfrom"])
            cal_tz_sub.add("tzoffsetto", tzinfo["tzoffsetto"])
            # TODO: add rrule
            # tzi.add('rrule',
            #         {'freq': 'yearly', 'bymonth': 10, 'byday': '-1su'})
            cal_tz.add_component(cal_tz_sub)
        cal.add_component(cal_tz)

    return cal


def add_to_zones_map(tzmap, tzid, dt):
    """Build a dictionary of timezone information from a timezone identifier
    and a date/time object for which the timezone information should be
    calculated.

    :param tzmap: An existing dictionary of timezone information to be extended
                  or an empty dictionary.
    :type tzmap: dictionary
    :param tzid: A timezone identifier.
    :type tzid: string
    :param dt: A datetime object.
    :type dt: datetime

    :returns: A dictionary with timezone information needed to build VTIMEZONE
              entries.
    :rtype: dictionary
    """

    if tzid.lower() == "utc" or not is_datetime(dt):
        # no need to define UTC nor timezones for date objects.
        return tzmap
    tz = pytz.timezone(tzid)
    transitions = getattr(tz, "_utc_transition_times", None)
    if not transitions:
        return tzmap  # we need transition definitions
    dtzl = tzdel(utc(dt))

    # get transition time, which is the dtstart of timezone.
    #     the key function returns the value to compare with. as long as item
    #     is smaller or equal like the dt value in UTC, return the item. as
    #     soon as it becomes greater, compare with the smallest possible
    #     datetime, which wouldn't create a match within the max-function. this
    #     way we get the maximum transition time which is smaller than the
    #     given datetime.
    transition = max(transitions, key=lambda item: item if item <= dtzl else dt.min)

    # get previous transition to calculate tzoffsetfrom
    idx = transitions.index(transition)
    prev_idx = idx - 1 if idx > 0 else idx
    prev_transition = transitions[prev_idx]

    def localize(tz, dt):
        if dt is datetime.min:
            # dummy time, edge case
            # (dt at beginning of all transitions)
            return dt.min
        return pytz.utc.localize(dt).astimezone(tz)  # naive to utc + localize

    transition = localize(tz, transition)
    dtstart = tzdel(transition)  # timezone dtstart must be in local time
    prev_transition = localize(tz, prev_transition)

    if tzid not in tzmap:
        tzmap[tzid] = {}  # initial
    if dtstart in tzmap[tzid]:
        return tzmap  # already there
    tzmap[tzid][dtstart] = {
        "dst": transition.dst() > timedelta(0),
        "name": transition.tzname(),
        "tzoffsetfrom": prev_transition.utcoffset(),
        "tzoffsetto": transition.utcoffset(),
        # TODO: recurrence rule
    }
    return tzmap


@implementer(IICalendar)
def calendar_from_event(context):
    """Event adapter. Returns an icalendar.Calendar object from an Event
    context.
    """
    context = aq_inner(context)
    return construct_icalendar(context, [context])


@implementer(IICalendar)
def calendar_from_container(context):
    """Container adapter. Returns an icalendar.Calendar object from a
    Containerish context like a Folder.
    """
    context = aq_inner(context)
    path = "/".join(context.getPhysicalPath())
    result = get_events(context, ret_mode=RET_MODE_BRAINS, expand=False, path=path)
    return construct_icalendar(context, result)


@implementer(IICalendar)
def calendar_from_collection(context):
    """Container/Event adapter. Returns an icalendar.Calendar object from a
    Collection.
    """
    context = aq_inner(context)
    # The keyword argument brains=False was added to plone.app.contenttypes
    # after 1.0
    result = context.results(batch=False, sort_on="start")
    return construct_icalendar(context, result)


@implementer(IICalendarEventComponent)
class ICalendarEventComponent:
    """Returns an icalendar object of the event."""

    def __init__(self, context):
        self.context = context
        self.event = IEventAccessor(self.context)
        self.ical = icalendar.Event()

    @property
    def dtstamp(self):
        # must be in uc
        return {"value": utc(datetime.now())}

    @property
    def created(self):
        # must be in uc
        return {"value": utc(self.event.created)}

    @property
    def last_modified(self):
        # must be in uc
        return {"value": utc(self.event.last_modified)}

    @property
    def uid(self):
        return {"value": self.event.sync_uid}

    @property
    def url(self):
        return {"value": self.event.url}

    @property
    def summary(self):
        return {"value": self.event.title}

    @property
    def description(self):
        return {"value": self.event.description}

    @property
    def dtstart(self):
        if self.event.whole_day:
            # RFC5545, 3.6.1
            # For cases where a "VEVENT" calendar component
            # specifies a "DTSTART" property with a DATE value type but no
            # "DTEND" nor "DURATION" property, the event's duration is taken to
            # be one day.
            return {"value": self.event.start.date()}

        # Normal case + Open End case:
        # RFC5545, 3.6.1
        # For cases where a "VEVENT" calendar component
        # specifies a "DTSTART" property with a DATE-TIME value type but no
        # "DTEND" property, the event ends on the same calendar date and
        # time of day specified by the "DTSTART" property.
        return {"value": self.event.start}

    @property
    def dtend(self):
        if self.event.whole_day:
            # RFC5545, 3.6.1
            # For cases where a "VEVENT" calendar component
            # specifies a "DTSTART" property with a DATE value type but no
            # "DTEND" nor "DURATION" property, the event's duration is taken to
            # be one day.
            #
            # RFC5545 doesn't define clearly, if all-day events should have
            # a end date on the same date or one day after the start day at
            # 0:00. Most icalendar libraries use the latter method.
            # Internally, whole_day events end on the same day one second
            # before midnight. Using the RFC5545 preferred method for
            # plone.app.event seems not appropriate, since we would have to fix
            # the date to end a day before for displaying.
            # For exporting, we let whole_day events end on the next day at 0:00.
            # See:
            # http://stackoverflow.com/questions/1716237/single-day-all-day-appointments-in-ics-files
            # http://icalevents.com/1778-all-day-events-adding-a-day-or-not/
            # http://www.innerjoin.org/iCalendar/all-day-events.html
            return {"value": self.event.end.date() + timedelta(days=1)}

        elif self.event.open_end:
            # RFC5545, 3.6.1
            # For cases where a "VEVENT" calendar component
            # specifies a "DTSTART" property with a DATE-TIME value type but no
            # "DTEND" property, the event ends on the same calendar date and
            # time of day specified by the "DTSTART" property.
            return None

        return {"value": self.event.end}

    @property
    def recurrence(self):
        if not self.event.recurrence or IOccurrence.providedBy(self.context):
            return None

        ret = []
        for recdef in self.event.recurrence.split():
            prop, val = recdef.split(":")
            if prop == "RRULE":
                ret.append(
                    {"property": prop, "value": icalendar.prop.vRecur.from_ical(val)}
                )

            elif prop in ("EXDATE", "RDATE"):
                factory = icalendar.prop.vDDDLists

                # localize ex/rdate
                # TODO: should better already be localized by event object
                tzid = self.event.timezone
                if isinstance(tzid, tuple):
                    tzid = tzid[0]
                # get list of datetime values from ical string
                try:
                    dtlist = factory.from_ical(val, timezone=tzid)
                except ValueError:
                    # TODO: caused by a bug in plone.formwidget.recurrence,
                    # where the recurrencewidget or plone.event fails with
                    # COUNT=1 and a extra RDATE.
                    # TODO: REMOVE this workaround, once this failure is
                    # fixed in recurrence widget.
                    continue
                ret.append({"property": prop, "value": dtlist})

        return ret

    @property
    def location(self):
        return {"value": self.event.location}

    @property
    def attendee(self):
        # TODO: revisit and implement attendee export according to RFC
        ret = []
        for attendee in self.event.attendees or []:
            att = icalendar.prop.vCalAddress(attendee)
            att.params["cn"] = icalendar.prop.vText(attendee)
            att.params["ROLE"] = icalendar.prop.vText("REQ-PARTICIPANT")
            ret.append(att)
        return {"value": ret}

    @property
    def contact(self):
        cn = []
        event = self.event
        if event.contact_name:
            cn.append(event.contact_name)
        if event.contact_phone:
            cn.append(event.contact_phone)
        if event.contact_email:
            cn.append(event.contact_email)
        if event.event_url:
            cn.append(event.event_url)

        return {"value": ", ".join(cn)}

    @property
    def categories(self):
        ret = []
        for cat in self.event.subjects or []:
            ret.append(cat)
        if ret:
            return {"value": ret}

    @property
    def geo(self):
        """Not implemented."""
        return

    def ical_add(self, prop, val):
        if not val:
            return

        if not isinstance(val, list):
            val = [val]

        for _val in val:
            assert isinstance(_val, dict)
            value = _val["value"]
            if not value:
                continue
            prop = _val.get("property", prop)
            params = _val.get("parameters", None)
            self.ical.add(prop, value, params)

    def to_ical(self):
        # TODO: event.text

        ical_add = self.ical_add
        ical_add("dtstamp", self.dtstamp)
        ical_add("created", self.created)
        ical_add("last-modified", self.last_modified)
        ical_add("uid", self.uid)
        ical_add("url", self.url)
        ical_add("summary", self.summary)
        ical_add("description", self.description)
        ical_add("dtstart", self.dtstart)
        ical_add("dtend", self.dtend)
        ical_add(None, self.recurrence)  # property key set via val
        ical_add("location", self.location)
        ical_add("attendee", self.attendee)
        ical_add("contact", self.contact)
        ical_add("categories", self.categories)
        ical_add("geo", self.geo)

        return self.ical


class EventsICal(BrowserView):
    """Returns events in iCal format."""

    def get_ical_string(self):
        cal = IICalendar(self.context)
        return cal.to_ical()

    def __call__(self):
        ical = self.get_ical_string()
        name = f"{self.context.getId()}.ics"
        self.request.response.setHeader("Content-Type", "text/calendar")
        self.request.response.setHeader(
            "Content-Disposition", f'attachment; filename="{name}"'
        )
        self.request.response.setHeader("Content-Length", len(ical))
        self.request.response.write(ical)



from DateTime import DateTime
from plone.app.event.base import localized_today
from plone.app.event.testing import make_fake_response
from plone.app.event.testing import PAEventDX_INTEGRATION_TESTING
from plone.app.event.tests.base_setup import AbstractSampleDataEvents
from plone.app.event.tests.base_setup import patched_now as PN
from unittest import mock


class TestEventsListingPortal(AbstractSampleDataEvents):
    layer = PAEventDX_INTEGRATION_TESTING

    def _listing_view(self, name="@@event_listing"):
        return self.portal.restrictedTraverse(name)

    @mock.patch("plone.app.event.browser.event_listing.localized_now", new=PN)
    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_get_events_future(self):
        # Default mode is to show all events from now on.
        view = self._listing_view()
        self.assertEqual(len(view.events(batch=False)), 5)

    @mock.patch("plone.app.event.browser.event_listing.localized_now", new=PN)
    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_get_events_past(self):
        self.request.form.update({"mode": "past"})
        view = self._listing_view()
        self.assertEqual(len(view.events(batch=False)), 5)

    @mock.patch("plone.app.event.browser.event_listing.localized_now", new=PN)
    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_get_events_all(self):
        self.request.form.update({"mode": "all"})
        view = self._listing_view()
        self.assertEqual(len(view.events(batch=False)), 8)

    @mock.patch("plone.app.event.browser.event_listing.localized_now", new=PN)
    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_get_events_today(self):
        today = localized_today().isoformat()
        self.request.form.update({"mode": "day", "date": today})
        view = self._listing_view()
        self.assertEqual(len(view.events(batch=False)), 2)

    @mock.patch("plone.app.event.browser.event_listing.localized_now", new=PN)
    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_events_listing_ical(self):
        # Default mode is to show all events from now on.
        headers, output, request = make_fake_response(self.request)
        view = self._listing_view(name="@@event_listing_ical")
        view()
        self.assertEqual(len(headers), 3)
        self.assertEqual(headers["Content-Type"], "text/calendar")
        self.assertTrue("Content-Length" in headers)
        icalstr = b"".join(output)
        self.assertTrue(b"Long Event" in icalstr)


class TestEventsListingCollection(TestEventsListingPortal):
    def _listing_view(self, name="@@event_listing"):
        return self.portal.collection.restrictedTraverse(name)

    def _collection_batching_base(self):
        """Base preparation for batching tests below.

        The event_listing view caches it's results on the request, so the
        batching tests below need to be in separate test methods to get a fresh
        environment with newly calculated results.
        """
        # plone.app.contenttypes ICollection type
        self.portal.invokeFactory("Collection", "col_test", title="Col")
        collection = self.portal.col_test
        collection.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.selection.any",
                "v": ["Event", "plone.app.event.dx.event"],
            },
        ]
        self.request.form.update({"mode": "all"})
        return collection

    def test_collection_batching__all(self):
        """Don't limit the results."""
        collection = self._collection_batching_base()
        view = collection.restrictedTraverse("@@event_listing")
        self.assertEqual(view.events(batch=True).length, 8)

    def test_collection_batching__reduce_by_collection_setting(self):
        """Limit the results by setting item_count on the collection."""
        collection = self._collection_batching_base()
        view = collection.restrictedTraverse("@@event_listing")
        collection.item_count = 4
        view = collection.restrictedTraverse("@@event_listing")
        self.assertEqual(view.events(batch=True).length, 4)

    def test_collection_batching__reduce_by_request_parameter(self):
        """Limit the results by using a request parameter."""
        collection = self._collection_batching_base()
        self.request.form.update({"b_size": 2})
        view = collection.restrictedTraverse("@@event_listing")
        self.assertEqual(view.events(batch=True).length, 2)

    def test_date_filtering(self):
        """Test if date filters are available on Collections without start or
        end search criteria.
        """
        # plone.app.contenttypes ICollection type
        self.portal.invokeFactory(
            "Collection", "col_without_date_criterion", title="Col"
        )
        collection = self.portal.col_without_date_criterion
        collection.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.selection.any",
                "v": ["Event", "plone.app.event.dx.event"],
            },
        ]
        view = self.portal.col_without_date_criterion.restrictedTraverse(
            "@@event_listing"
        )
        out = view()
        self.assertTrue("mode_past" in out)

    def test_no_date_filtering(self):
        """Test if date filters are not available on Collections with start or
        end search criteria.
        """
        # plone.app.contenttypes ICollection type
        self.portal.invokeFactory("Collection", "col_with_date_criterion", title="Col")
        collection = self.portal.col_with_date_criterion
        collection.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.selection.any",
                "v": ["Event", "plone.app.event.dx.event"],
            },
            {
                "i": "start",
                "o": "plone.app.querystring.operation.date.afterToday",
                "v": "",
            },
        ]
        view = self.portal.col_with_date_criterion.restrictedTraverse("@@event_listing")
        out = view()
        for _class in [
            "mode_future",
            "mode_past",
            "mode_month",
            "mode_week",
            "mode_day",
        ]:
            self.assertTrue(_class not in out)

    @mock.patch("plone.app.event.browser.event_listing.localized_now", new=PN)
    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_end_date_filtering(self):
        """Test date conversion between collection query and 'expand_events'
        filtering.
        """
        # plone.app.contenttypes ICollection type
        self.portal.invokeFactory("Collection", "col_with_date_criterion", title="Col")
        collection = self.portal.col_with_date_criterion
        collection.query = [
            {
                "i": "portal_type",
                "o": "plone.app.querystring.operation.selection.any",
                "v": ["Event", "plone.app.event.dx.event"],
            },
            {
                "i": "end",
                "o": "plone.app.querystring.operation.date.largerThan",
                "v": DateTime(PN()),
            },
        ]

        view = self.portal.col_with_date_criterion.restrictedTraverse("@@event_listing")
        view()

        self.assertIs(len(self.portal.col_with_date_criterion.results()), 3)



from plone.event.interfaces import IEvent
from plone.event.interfaces import IEventRecurrence


class IDXEvent(IEvent):
    """Marker interface for Dexterity events."""


class IDXEventRecurrence(IEventRecurrence):
    """Marker interface for recurring Dexterity events."""



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


version = "5.1.6"


long_description = "\n\n".join(
    [
        Path("README.rst").read_text(),
        (Path(".") / "docs" / "installation.rst").read_text(),
        Path("CHANGES.rst").read_text(),
    ]
)


setup(
    name="plone.app.event",
    version=version,
    description="The Plone calendar framework",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="plone event",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://github.com/plone/plone.app.event",
    license="GPL version 2",
    packages=find_packages(),
    namespace_packages=["plone", "plone.app"],
    include_package_data=True,
    python_requires=">=3.8",
    zip_safe=False,
    install_requires=[
        "setuptools",
        "plone.base",
        "Products.statusmessages",
        "icalendar",  # >4.0.2
        "Products.DateRecurringIndex",
        "Products.ZCatalog",
        "Products.GenericSetup",
        "plone.app.contenttypes",
        "plone.app.uuid",
        "plone.resource",
        "plone.app.contentlisting",
        "plone.app.dexterity",
        "plone.app.portlets >= 2.5.1",
        "plone.app.querystring",
        "plone.app.registry",
        "plone.app.textfield",
        "plone.app.vocabularies >= 2.1.15",
        "plone.app.z3cform>=2.0.1",
        "plone.autoform>=1.4",
        "plone.behavior",
        "plone.browserlayer",
        "plone.dexterity",
        "plone.event>=1.0",
        "plone.folder",
        "plone.formwidget.recurrence [z3cform] >= 1.2.4",
        "plone.indexer",
        "plone.memoize",
        "plone.namedfile",
        "plone.portlets",
        "plone.supermodel",
        "plone.uuid",
        "pytz",
        "z3c.form >= 3.2.1",
        "Zope",
    ],
    extras_require={
        "test": [
            "plone.app.robotframework",
            "plone.app.testing [robot]",
            "plone.testing",
            "robotsuite",
            "lxml",
        ],
    },
)



from plone.app.event.testing import PAEventDX_FUNCTIONAL_TESTING
from plone.app.event.tests.base_setup import AbstractSampleDataEvents
from plone.app.event.tests.base_setup import patched_now as PN
from unittest import mock


class FunctionalTestSearchEvent(AbstractSampleDataEvents):
    layer = PAEventDX_FUNCTIONAL_TESTING

    @mock.patch("plone.app.event.base.localized_now", new=PN)
    def test_searchabletext(self):
        results = self.portal.portal_catalog(SearchableText="Überraschung")
        self.assertTrue(len(results) == 1)



from Acquisition import aq_parent
from plone.app.event import _
from plone.event.interfaces import IEventAccessor
from plone.event.interfaces import IOccurrence
from plone.event.interfaces import IRecurrenceSupport
from plone.memoize import view
from plone.uuid.interfaces import IUUID
from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from zope.component import getMultiAdapter
from zope.contentprovider.interfaces import IContentProvider


class EventSummaryView(BrowserView):
    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.data = IEventAccessor(context)
        self.max_occurrences = 6
        self.excludes = [
            "title",
        ]

    @property
    def is_occurrence(self):
        return IOccurrence.providedBy(self.context)

    @property
    def event_context(self):
        if self.is_occurrence:
            return aq_parent(self.context)
        return self.context

    def formatted_date(self, occ):
        provider = getMultiAdapter(
            (self.context, self.request, self), IContentProvider, name="formatted_date"
        )
        return provider(occ)

    @property
    @view.memoize
    def next_occurrences(self):
        """Returns occurrences for this context, except the start
        occurrence, limited to self.max_occurrence occurrences.

        :returns: List with next occurrences.
        :rtype: list
        """
        occurrences = []
        adapter = IRecurrenceSupport(self.event_context, None)
        if adapter:
            for cnt, occ in enumerate(adapter.occurrences(range_start=self.data.start)):
                if cnt == self.max_occurrences:
                    break
                occurrences.append(occ)
        return occurrences

    @property
    @view.memoize
    def num_more_occurrences(self):
        """Return the number of extra occurrences, which are not listed by
        next_occurrences.
        """
        uid = IUUID(self.event_context, None)
        if not uid:
            # Might be an occurrence
            return 0
        catalog = getToolByName(self.event_context, "portal_catalog")
        brains = catalog(UID=uid)
        if len(brains) == 0:
            return 0
        brain = brains[0]  # assuming, that current context is in the catalog
        idx = catalog.getIndexDataForRID(brain.getRID())
        num = len(idx["start"]) - len(self.next_occurrences)
        return num if num > 0 else 0

    @property
    def more_occurrences_text(self):
        msgid = _(
            "msg_num_more_occurrences",
            default="There are ${results} more occurrences.",
            mapping={"results": self.num_more_occurrences},
        )
        return self.context.translate(msgid)

    @property
    @view.memoize
    def has_occurrences(self):
        occs = [o for o in self.next_occurrences if IOccurrence.providedBy(o)]
        return len(occs)
